import os, sys, glob, re

from tlfp.tools.common import strip_text, compute_similarity_by_articles, open_json, print_json, \
    clean_text_for_diff, datize, clean_statuses


def count_words(content):
    """
    Counts the number of words in the specified string.
    Based on the regexes / logic from Countable.js:
    https://github.com/RadLikeWhoa/Countable

    Taken from: https://github.com/Jaza/word-count/blob/master/word_count.py
    With fix from https://github.com/Jaza/word-count/issues/1
    And updated regex from: https://github.com/RadLikeWhoa/Countable/blob/4d2812d7b53736d2bca4268b783997545d261c43/Countable.js#L149

    Divergences with others applications:
        ╔════════════════════════╦═══════════════════════════════════════════════════════════════════════════╗
        ║                        ║                               Number of words                             ║
        ║ Input                  ╠═════════════╦══════════════╦═════════════╦════════════════╦═══════╦═══════╣
        ║                        ║ This script ║ Libre Office ║ Google Docs ║ Microsoft Word ║ Gedit ║ wc -w ║
        ╠════════════════════════╬═════════════╬══════════════╬═════════════╬════════════════╬═══════╬═══════╣
        ║ "L'article"            ║ 2           ║ 1            ║ 1           ║ 1              ║ 2     ║ 1     ║
        ╠════════════════════════╬═════════════╬══════════════╬═════════════╬════════════════╬═══════╬═══════╣
        ║ "Ceci :"               ║ 1           ║ 2            ║ 1           ║ 2              ║ 1     ║ 2     ║
        ╠════════════════════════╬═════════════╬══════════════╬═════════════╬════════════════╬═══════╬═══════╣
        ║ "321-32"               ║ 1           ║ 2            ║ 1           ║ 1              ║ 2     ║ 1     ║
        ╠════════════════════════╬═════════════╬══════════════╬═════════════╬════════════════╬═══════╬═══════╣
        ║ "L'article L.O. 321-3" ║ 4           ║ 3            ║ 4           ║ 3              ║ 6     ║ 3     ║
        ╚════════════════════════╩═════════════╩══════════════╩═════════════╩════════════════╩═══════╩═══════╝

    NOTE: "wc -w" only count spaces-separated words

    >>> count_words("L'article L.O. 321-3")
    4
    >>> count_words("<td>Exécution 2017</td><td>Prévision d'exécution 2018</td>")
    6
    """

    striptags_re = re.compile(r'<\/?[a-z][^>]*>', re.IGNORECASE)
    stripsymbols_re = re.compile(r'[";:,.?¿\-!¡/]+')
    splitsymbols_re = re.compile(r"'")
    words_re = re.compile(r'\S+')

    c = content.strip()
    c = striptags_re.sub(' ', c)
    c = stripsymbols_re.sub('', c)
    c = splitsymbols_re.sub(' ', c)

    match = words_re.findall(c)

    return len(match) if match else 0


def has_been_censored(dos):
    cc_step = [step for step in dos['steps'] if step.get('stage') == 'constitutionnalité']
    if cc_step:
        return cc_step[0].get('decision') == 'partiellement conforme'


def find_amendements(path):
    for amdts_file in glob.glob(os.path.join(path, '**/amendements_*'), recursive=True):
        amendements = open_json(amdts_file)
        for subject in amendements.get('sujets', {}).values():
            for amdt in subject.get('amendements', []):
                yield amdt, amdts_file


def read_alineas(art):
    return [art['alineas'][al] for al in sorted(art['alineas'].keys())]


def read_text(step):
    articles = step['texte.json']['articles']
    for art in articles:
        # yield art['titre']
        for al in read_alineas(art):
            yield al


def step_word_count(step):
    return count_words('\n'.join(clean_statuses(al) for al in read_text(step)))


def step_text_length(step):
    return len(''.join(strip_text(al) for al in read_text(step)))


def count_initial_depots(steps):
    count = 0
    for step in steps:
        if step.get('step') == 'depot':
            count += 1
        else:
            break
    return count


def count_navettes(steps):
    count = 0
    last_step = None
    for step in steps:
        if step.get('step') == 'depot':
            if not last_step or last_step.get('step') != 'depot':
                count += 1
        elif step.get('stage') == 'CMP' and step.get('step') == 'commission':
            count += 1
        elif step.get('stage') in ('congrès', 'constitutionnalité'):
            count += 1
        last_step = step
    return count


def count_texts(steps):
    return len([step for step in steps if step.get('debats_order') is not None])


def read_articles(step):
    articles = step['texte.json']['articles']
    arts = {art['titre']: clean_text_for_diff(read_alineas(art)) for art in articles}
    return {titre: txt for titre, txt in arts.items() if txt}


def count_censored_articles(step):
    articles = step['texte.json']['articles']
    censored_articles = 0
    fully_censored_articles = 0
    for art in articles:
        txt = ''.join(read_alineas(art))
        if '(Censuré)' in txt:
            censored_articles += 1
        if txt == '(Censuré)':
            fully_censored_articles += 1
    return censored_articles, fully_censored_articles


def find_first_and_last_steps(dos, include_CC=True):
    first, last = None, None
    first_found = False
    for i, s in enumerate(dos['steps']):
        if s['debats_order'] is None or s.get('echec') or s.get('in_discussion'):
            continue
        if s.get('stage') == 'constitutionnalité' and not include_CC:
            break
        if s.get('step') != "depot":
            first_found = True
            last = i
        if not first_found and s.get('step') == "depot":
            first = last = i
    return dos['steps'][first], dos['steps'][last]


def add_amendments_stats(stats, amendements):
    stats['total_amendements'] \
        = stats["total_amendements_adoptes"] \
        = stats["total_amendements_senateurs"] \
        = stats["total_amendements_senateurs_adoptes"] \
        = stats["total_amendements_gouvernement"] \
        = stats["total_amendements_gouvernement_adoptes"] \
        = stats["total_amendements_gouvernement_senat"] \
        = stats["total_amendements_gouvernement_senat_adoptes"] \
        = stats["total_amendements_deputes"] \
        = stats["total_amendements_deputes_adoptes"] \
        = stats["total_amendements_gouvernement_assemblee"] \
        = stats["total_amendements_gouvernement_assemblee_adoptes"] \
        = stats['total_amendements_hemicycle'] \
        = stats["total_amendements_hemicycle_adoptes"] \
        = stats["total_amendements_hemicycle_senateurs"] \
        = stats["total_amendements_hemicycle_senateurs_adoptes"] \
        = stats["total_amendements_hemicycle_gouvernement"] \
        = stats["total_amendements_hemicycle_gouvernement_adoptes"] \
        = stats["total_amendements_hemicycle_gouvernement_senat"] \
        = stats["total_amendements_hemicycle_gouvernement_senat_adoptes"] \
        = stats["total_amendements_hemicycle_deputes"] \
        = stats["total_amendements_hemicycle_deputes_adoptes"] \
        = stats["total_amendements_hemicycle_gouvernement_assemblee"] \
        = stats["total_amendements_hemicycle_gouvernement_assemblee_adoptes"] \
        = 0

    for amdt, amdts_file in amendements:
        senat = '_senat' in amdts_file
        assemblee = '_assemblee' in amdts_file
        hemicycle = '_hemicycle' in amdts_file
        from_gouv = amdt["groupe"] == "Gouvernement"

        stats['total_amendements'] += 1
        if hemicycle:
            stats['total_amendements_hemicycle'] += 1

        if amdt["sort"] == "adopté":
            stats["total_amendements_adoptes"] += 1
            if hemicycle:
                stats["total_amendements_hemicycle_adoptes"] += 1

            if from_gouv:
                stats["total_amendements_gouvernement_adoptes"] += 1
                if hemicycle:
                    stats["total_amendements_hemicycle_gouvernement_adoptes"] += 1

            if senat:
                stats["total_amendements_senateurs_adoptes"] += 1
                if hemicycle:
                    stats["total_amendements_hemicycle_senateurs_adoptes"] += 1

                if from_gouv:
                    stats["total_amendements_gouvernement_senat_adoptes"] += 1
                    if hemicycle:
                        stats["total_amendements_hemicycle_gouvernement_senat_adoptes"] += 1

            if assemblee:
                stats["total_amendements_deputes_adoptes"] += 1
                if hemicycle:
                    stats["total_amendements_hemicycle_deputes_adoptes"] += 1

                if from_gouv:
                    stats["total_amendements_gouvernement_assemblee_adoptes"] += 1
                    if hemicycle:
                        stats["total_amendements_hemicycle_gouvernement_assemblee_adoptes"] += 1

        if from_gouv:
            stats["total_amendements_gouvernement"] += 1
            if hemicycle:
                stats["total_amendements_hemicycle_gouvernement"] += 1

        if senat:
            stats["total_amendements_senateurs"] += 1
            if hemicycle:
                stats["total_amendements_hemicycle_senateurs"] += 1

            if from_gouv:
                stats["total_amendements_gouvernement_senat"] += 1
                if hemicycle:
                    stats["total_amendements_hemicycle_gouvernement_senat"] += 1

        if assemblee:
            stats["total_amendements_deputes"] += 1
            if hemicycle:
                stats["total_amendements_hemicycle_deputes"] += 1

            if from_gouv:
                stats["total_amendements_gouvernement_assemblee"] += 1
                if hemicycle:
                    stats["total_amendements_hemicycle_gouvernement_assemblee"] += 1


def process(output_dir, dos):
    stats = {}

    # # # INTERVENTIONS # # #

    intervs = open_json(os.path.join(output_dir, 'viz/interventions.json'))
    # only keep seances in hemicycle
    intervs = {step_name: step for step_name, step in intervs.items() if '_hemicycle' in step_name}

    stats['total_mots'] = sum([
        sum(i['total_mots'] for i in step['divisions'].values())
            for step in intervs.values()
    ])

    stats["total_intervenants"] = len({orat for step in intervs.values() for orat in step['orateurs'].keys()})
    stats["total_interventions"] = sum({division['total_intervs'] for step in intervs.values() for division in step['divisions'].values()})

    stats["total_seances"] = sum([step['total_seances'] for step in intervs.values()])
    stats["total_seances_assemblee"] = sum([step['total_seances'] for dir, step in intervs.items() if '_assemblee' in dir])
    stats["total_seances_senat"] = sum([step['total_seances'] for dir, step in intervs.items() if '_senat' in dir])

    # # # AMENDMENTS # # #

    add_amendments_stats(stats, find_amendements(output_dir))

    # # # TEXTS # # #

    first_step, last_step = find_first_and_last_steps(dos)
    first_arts = read_articles(first_step)
    last_arts = read_articles(last_step)

    stats["total_input_articles"] = len(first_arts)
    stats["total_output_articles"] = len(last_arts)
    stats["ratio_articles_growth"] = (stats["total_output_articles"] - stats["total_input_articles"]) / stats["total_input_articles"]

    stats["input_text_length"] = step_text_length(first_step)
    stats["output_text_length"] = step_text_length(last_step)
    stats["ratio_text_length_growth"] = (stats["output_text_length"] - stats["input_text_length"]) / stats["input_text_length"]

    stats["input_text_word_count"] = step_word_count(first_step)
    stats["output_text_word_count"] = step_word_count(last_step)
    stats["ratio_word_count_growth"] = (stats["output_text_word_count"] - stats["input_text_word_count"]) / stats["input_text_word_count"]

    adopted_step = find_first_and_last_steps(dos, include_CC=False)[1]
    if has_been_censored(dos):
        stats["censored_articles"], stats["fully_censored_articles"] = count_censored_articles(last_step)
        stats["output_text_before_CC_length"] = step_text_length(adopted_step)
        stats["output_text_before_CC_word_count"] = step_word_count(adopted_step)

    stats["ratio_texte_modif"] = 1 - compute_similarity_by_articles(first_arts, last_arts)

    # # # PROCEDURE # # #

    stats["echecs_procedure"] = len([step for step in dos['steps'] if step.get("echec")])

    # TODO: first institution
    stats['last_stage'] = adopted_step.get('stage')
    if stats['last_stage'] == 'CMP':
        stats['last_institution'] = 'CMP'
    else:
        stats['last_institution'] = adopted_step.get('institution')

    maxdate = dos.get('end')
    if not maxdate:
        for step in dos['steps']:
            if step.get('date'):
                maxdate = step.get('enddate') or step.get('date')
    stats["total_days"] = (datize(maxdate) - datize(dos['beginning'])).days + 1

    stats["attached_law_proposals"] = count_initial_depots(dos['steps']) - 1
    stats["depots_in_institutions"] = count_navettes(dos['steps'])
    stats["texts_produced"] = count_texts(dos['steps'])

    return stats


if __name__ == '__main__':
    print_json(process(sys.argv[1], open_json(sys.argv[2])))



#!/usr/bin/python
# -*- coding=utf-8 -*-
"""Create file arborescence corresponding to a law json output from parse_texte.py

Run with python json2arbo.py JSON_FILE PROJECT
where LAW_FILE results from perl download_loi.pl URL > LAW_FILE
Outputs results to stdout"""

import os, sys, re
import urllib.parse

from .common import print_json, open_json


def mkdirs(d):
    if not os.path.exists(d):
        os.makedirs(d)


def get_step_id(nstep, step):
    clean = lambda x: x.replace(' ', '').replace('è','e').lower() if x else ''
    return '%s_%s_%s_%s' % (
        str(nstep).zfill(2),
        clean(step.get('stage')),
        clean(step.get('institution')),
        clean(step.get('step')),
    )


def process(dos, OUTPUT_DIR):
    def log_err(txt):
        raise Exception(txt)

    for step_i, step in enumerate(dos['steps']):

        step['directory'] = get_step_id(step_i, step)
        step_dir = os.path.join(OUTPUT_DIR, os.path.join(step['directory'], 'texte'))

        articles = step.get('articles_completed', step.get('articles'))
        if not articles:
            continue

        mkdirs(step_dir)
        textid = None
        for data in articles:
            if not data or not "type" in data:
                log_err("JSON badly formatted, missing field type: %s" % data)
            if data["type"] == "texte":
                textid = data["id"]
                alldata = dict(data)
                alldata['sections'] = []
                alldata['articles'] = []
            elif not textid:
                log_err("JSON missing first line with text infos for step %s" % step_dir)
            elif data["type"] == "section":
                alldata['sections'].append(data)
            elif data["type"] == "article":
                alldata['articles'].append(data)
            elif data["type"] == "echec":
                alldata['expose'] = data['texte']


        print_json(alldata, os.path.join(step_dir, 'texte.json'))

        step['texte.json'] = alldata

    return dos


if __name__ == '__main__':
    print_json(process(open_json(sys.argv[1]), 'test_out'))



import os, shutil, io

from .tools import json2arbo, prepare_articles, update_procedure, \
    prepare_amendements, prepare_interventions, reorder_interventions_and_correct_procedure, \
    compute_stats, add_links
from .tools.common import debug_file, print_json


def project_header_template(procedure):
    return """
<h1>Les données pour: "{long_title}"</h1>
<p>Les données mises à disposition dans ces répertoires sont celles utilisées par <a href="https://lafabriquedelaloi.fr/">La Fabrique de la Loi</a> pour visualiser "<a href="https://lafabriquedelaloi.fr/lois.html?loi={dos_id}">{long_title}</a>".</p>
<p>Elles ont été constituées par <a href="https://regardscitoyens.org">Regards Citoyens</a> à partir de <a href="https://nosdeputes.Fr/">NosDéputés.fr</a>, <a href="https://NosSénateurs.fr">NosSénateurs.fr<a/> et les sites du <a href="https://www.senat.fr/">Sénat</a> et de l'<a href="http://www.assemblee-nationale.fr">Assemblée nationale</a>. Elles sont réutilisables librement en <img src="https://www.nosdeputes.fr/images/opendata.png" alt="Open Data"/> sous la licence <a href="https://opendatacommons.org/licenses/odbl/">ODBL</a>.</p>
<p>Le répertoire <a href="procedure/"><img src="/icons/folder.gif"/>&nbsp;procedure/</a> contient un dossier par étape de la procédure parlementaire pour ce projet, avec pour chacun lorsqu'elle existe la version finale du texte de loi à cette étape. Le répertoire <a href="viz/"><img src="/icons/folder.gif"/>&nbsp;viz/</a> contient les fichiers utilisés par l'application.</p>
""".format(long_title=procedure.get('long_title'), dos_id=procedure['id'])


def dump_success_log(output_dir, log):
    log = log.getvalue()
    logfile = os.path.join(output_dir, "parsing.log")
    with open(logfile, 'w') as f:
        f.write(log)
    textid = output_dir.split('/')[-1]
    api_dir = output_dir.replace('/' + textid, '')

    for err_dir in ('logs', 'logs-encours'):
        err_log = os.path.join(api_dir, err_dir, textid)
        if os.path.exists(err_log):
            os.remove(err_log)


def process(dos, OUTPUT_DIR, log=io.StringIO()):
    dos['id'] = dos.get('senat_id', dos.get('assemblee_id'))

    output_dir = os.path.join(OUTPUT_DIR, dos['id'] + '_tmp')
    final_output_dir = os.path.join(OUTPUT_DIR, dos['id'])
    print('     writing to:', dos['id'] + '_tmp')

    shutil.rmtree(output_dir, ignore_errors=True)

    # avoid duplicate titles
    if 'short_title' in dos:
        if " de loi organique" in dos['long_title']:
            dos['short_title'] += " (texte organique)"
    # AN doslegs have no short_titles
    else:
        dos['short_title'] = dos['long_title']

    debug_file(dos, 'before_add_links.json')
    dos = add_links.process(dos)

    # add texte.json and write all the text files tree
    debug_file(dos, 'before_json2arbo.json')
    dos = json2arbo.process(dos, output_dir + '/procedure')

    print(' - process article versions')
    json2arbo.mkdirs(os.path.join(output_dir, 'viz'))
    debug_file(dos, 'before_prepare_articles.json')
    articles_etapes = prepare_articles.process(dos)
    print_json(articles_etapes, os.path.join(output_dir, 'viz', 'articles_etapes.json'))

    procedure = update_procedure.process(dos, articles_etapes)

    print(' - process amendements & interventions')
    procedure = prepare_amendements.process(output_dir, procedure)

    print(' - re-order interventions and correct procedure dates')
    procedure = reorder_interventions_and_correct_procedure.process(output_dir, procedure)

    print(' - prepare interventions.json')
    prepare_interventions.process(output_dir, procedure)

    print(' - compute stats')
    debug_file(dos, 'before_stats.json')
    procedure['stats'] = compute_stats.process(output_dir, procedure)

    # remove temporary data
    for step in procedure['steps']:
        for key in 'articles_completed', 'articles', 'texte.json':
            try:
                step.pop(key)
            except KeyError:
                pass

    # remove temporary interventions files
    for step in procedure['steps']:
        step_directory = os.path.join(output_dir, 'procedure', step['directory'])
        shutil.rmtree(os.path.join(step_directory, 'interventions'), ignore_errors=True)
        # remove empty step directory
        if os.path.exists(step_directory) and not os.listdir(step_directory):
            os.rmdir(step_directory)

    print_json(procedure, os.path.join(output_dir, 'viz', 'procedure.json'))

    with open(os.path.join(output_dir, 'HEADER.html'), 'w') as f:
        f.write(project_header_template(procedure))

    if 'assemblee_id' in dos:
        an_output_dir = os.path.join(OUTPUT_DIR, dos['assemblee_id'])
        shutil.rmtree(an_output_dir, ignore_errors=True)
    shutil.rmtree(final_output_dir, ignore_errors=True)
    os.rename(output_dir, final_output_dir)

    print('  FINISHED -', dos['id'])

    dump_success_log(final_output_dir, log)



#!/usr/bin/python

# Add links in texte

import sys, re
import metslesliens

from .common import open_json, print_json

STATS_BLACKLIST = [
    'loi',
    'ordonnance',
    'décret',
    'arrété',
    'circulaire',
    'directive',
    'règlement'
]


def get_code(candidat):
    code = candidat['texte']['nom'].lower()
    for prefix in STATS_BLACKLIST:
        if code.startswith(prefix):
            return None

    if 'numero' in candidat['texte']:
        code += ' ' + candidat['texte']['numero']
    elif 'date' in candidat['texte']:
        code += ' du ' + candidat['texte']['date']

    return code


def process(dos):
    for step_i, step in enumerate(dos['steps']):
        articles = step.get('articles_completed', step.get('articles'))
        if not articles:
            continue
        dos["textes_cites"] = [] # only keep the latest version
        for data in articles:
            if data["type"] == "article":
                data['liens'] = []
                for i in range(len(data["alineas"])):
                    text = data["alineas"]["%03d" % (i+1)]
                    for candidat in metslesliens.donnelescandidats(text, 'structuré'):
                        if 'texte' in candidat and 'relatif' not in candidat['texte']:
                            code = get_code(candidat)
                            if code and code not in dos["textes_cites"]:
                                dos["textes_cites"].append(code)

                            link_text = text[candidat['index'][0]:candidat['index'][1]]
                            link_text = re.sub(r'^(aux?|les?|la|du|des)(dite?s?)? ', '', link_text, 0, re.I)
                            link_text = re.sub(r"^l'", '', link_text, 0, re.I)
                            if not re.search(r'(même|présente?|précédente?) ', link_text ) and link_text not in data['liens']:
                                """
                                link = {
                                    'texte': link_text
                                }
                                if 'eli_alias' in candidat:
                                    link['url'] = "https://www.legifrance.gouv.fr/" + candidat['eli_alias']
                                """
                                data['liens'].append(link_text)
        dos["textes_cites"].sort()
    return dos


if __name__ == '__main__':
    print_json(process(open_json(sys.argv[1])))



import sys, io, os, traceback

from senapy.dosleg.parser import parse as senapy_parse
from anpy.dossier_like_senapy import parse as anpy_parse
from lawfactory_utils.urls import download, enable_requests_cache, parse_national_assembly_url

from . import format_data_for_frontend
from . import parse_doslegs_texts
from .tools.detect_anomalies import find_anomalies
from .tools.json2arbo import mkdirs
from .tools.download_groupes import process as download_groupes
from .tools.download_lois_dites import process as download_lois_dites
from .tools.download_AN_opendata import process as download_AN_opendata
from .tools.common import debug_file, log_print
from .merge import merge_senat_with_an


class ParsingFailedException(Exception):
    def __init__(self, exception, logfile):
        super().__init__()
        self.root_exception = exception
        self.logfile = logfile


def download_senat(url, log=sys.stderr):
    print('  [] download SENAT version')
    resp = download(url)
    if resp.status_code != 200:
        print('WARNING: Invalid response -', resp.status_code)
        return
    html = resp.text
    print('  [] parse SENAT version')
    senat_dos = senapy_parse(html, url, logfile=log)
    debug_file(senat_dos, 'senat_dos.json')
    return senat_dos


def download_an(url, cached_opendata_an, url_senat=False, log=sys.stderr):
    print('  [] download AN version')
    print('  [] parse AN version')
    # TODO: do both instead of first
    results = anpy_parse(url, logfile=log, cached_opendata_an=cached_opendata_an)
    if not results:
        print('     WARNING: AN DOS NOT FOUND', url)
        return
    an_dos = results[0]
    if len(results) > 1:
        if url_senat:
            for result in results:
                if result.get('url_dossier_senat') == url_senat:
                    an_dos = result
                    break
        print('     WARNING: TOOK FIRST DOSLEG BUT THERE ARE %d OF THEM' % len(results))

    debug_file(an_dos, 'an_dos.json')
    return an_dos


def are_same_doslegs(senat_dos, an_dos):
    # same dosleg url ?
    if an_dos['url_dossier_senat'] == senat_dos['url_dossier_senat']:
        return True
    elif download(an_dos['url_dossier_senat']).status_code == 404:
        return True
    # same first text  ?
    if senat_dos.get('steps') and an_dos.get('steps') \
        and senat_dos['steps'][0].get('source_url') == an_dos['steps'][0].get('source_url'):
        return True
    # it's not the same dosleg !
    return False


def download_merged_dos(url, cached_opendata_an, log=sys.stderr):
    """find dossier from url and returns (the_merged_dosleg, AN_dosleg, SENAT_dosleg)"""
    if not url.startswith('http') and ('pjl' in url or 'ppl' in url or 'plfss' in url):
        url = "http://www.senat.fr/dossier-legislatif/%s.html" % url

    print(' -= DOSLEG URL:', url, '=-')

    dos = None
    an_dos = None
    senat_dos = None

    if 'senat.fr' in url:
        senat_dos = download_senat(url, log=log)
        if not senat_dos:
            print('  /!\ INVALID SENAT DOS')
            return None, None, None
        # Add AN version if there's one
        if 'url_dossier_assemblee' in senat_dos:
            an_dos = download_an(senat_dos['url_dossier_assemblee'], cached_opendata_an, senat_dos['url_dossier_senat'], log=log)
            if not an_dos:
                return senat_dos, None, senat_dos
            if 'url_dossier_senat' in an_dos:
                assert are_same_doslegs(senat_dos, an_dos)
            dos = merge_senat_with_an(senat_dos, an_dos)
        else:
            dos = senat_dos
    elif 'assemblee-nationale.fr' in url:
        dos = an_dos = download_an(url, cached_opendata_an, log=log)
        # Add senat version if there's one
        if an_dos and 'url_dossier_senat' in an_dos:
            senat_dos = download_senat(an_dos['url_dossier_senat'], log=log)
            if senat_dos:
                dos = merge_senat_with_an(senat_dos, an_dos)
    else:
        print(' INVALID URL:', url)
    return dos, an_dos, senat_dos


def dump_error_log(url, exception, api_dir, logdir, log):
    url_id = url.replace('/', '')
    if 'assemblee-nationale' in url:
        url_id = "%s-%s" % parse_national_assembly_url(url)
    elif 'senat.fr' in url:
        url_id = url.split('/')[-1].replace('.html', '')

    abs_dir = os.path.join(api_dir, logdir)
    mkdirs(abs_dir)
    abs_file = os.path.join(abs_dir, url_id)

    with open(abs_file, 'w') as f:
        f.write(log.getvalue())

    print('[error] parsing of', url, 'failed. Details in', abs_file)

    raise ParsingFailedException(exception, os.path.join(logdir, url_id))


def process(API_DIRECTORY, url):
    only_promulgated = '--only-promulgated' in sys.argv
    quiet = '--quiet' in sys.argv
    if '--enable-cache' in sys.argv:
        enable_requests_cache()

    dos = None
    with log_print(only_log=quiet) as log:
        try:
            print('======')
            print(url)

            # download the AN open data or just retrieve the last stored version
            opendata_an = download_AN_opendata(API_DIRECTORY)

            dos, an_dos, senat_dos = download_merged_dos(url, opendata_an, log=log)
            if not dos:
                raise Exception('Nothing found at %s' % url)

            find_anomalies([dos])

            if not dos.get('url_jo') and only_promulgated:
                print('    ----- passed: no JO link')
                return

            print('        title:', dos.get('long_title'))

            debug_file(dos, 'dos.json')

            # download the groupes in case they are not there yet
            download_groupes(API_DIRECTORY)

            # Add potential common name from Legifrance's "Lois dites"
            common_laws = download_lois_dites(API_DIRECTORY)
            if dos.get('legifrance_cidTexte') in common_laws and common_laws[dos['legifrance_cidTexte']].lower() not in dos['short_title'].lower():
                dos['loi_dite'] = common_laws[dos['legifrance_cidTexte']]

            print('  [] parse the texts')
            dos_with_texts = parse_doslegs_texts.process(dos)

            print('  [] format data for the frontend')
            format_data_for_frontend.process(dos_with_texts, API_DIRECTORY, log=log)
            return dos
        except KeyboardInterrupt as e:  # bypass the error log dump when doing Ctrl-C
            raise e
        except Exception as e:
            print(*traceback.format_tb(e.__traceback__), e, sep='', file=log)
            # dump log for each failed doslegs in logs/
            logdir = 'logs'
            if dos and not dos.get('url_jo'):
                logdir = 'logs-encours'
            dump_error_log(url, e, API_DIRECTORY, logdir, log)


if __name__ == '__main__':
    args = [arg for arg in sys.argv[1:] if '--' not in arg]
    url = args[0]
    API_DIRECTORY = args[1] if len(args) > 1 else 'data'
    process(API_DIRECTORY, url)




def use_old_procedure(step, dos=None):
    if dos and dos.get('use_old_procedure'):
        return True
    return (step.get("enddate", step.get("date", "9999-99-99")) or "9999-99-99") < "2009-03-01"


def should_ignore_commission_text(step, dos):
    return step.get('step') == 'commission' and (
        step.get('stage') == 'l. définitive' or (
        use_old_procedure(step, dos) and step['institution'] in ('senat', 'assemblee'))
    )


def is_one_of_the_initial_depots(steps, step_index):
    # Detect if the step is one of the multiple initial depot
    return step_index == 0 or (steps[step_index - 1].get('step') == steps[step_index].get('step') == 'depot')


def get_previous_step(steps, curr_step_index, is_old_procedure=False, get_depot_step=False):
    # is_old_procedure: Budget, Financement Sécurité Sociale, lois organique
    # get_depot_step: get the step the amendments are done on (including the depot steps)

    curr_step = steps[curr_step_index]

    if curr_step.get('stage') == 'l. définitive' and (
            (curr_step.get('step') == 'hemicycle' and not get_depot_step) or
            curr_step.get('step') == 'depot'
        ):
        # cf Constitution Article 45 alinéa 4 https://www.legifrance.gouv.fr/affichTexteArticle.do?idArticle=LEGIARTI000006527521&dateTexte=&categorieLien=cid
        print('[step_logic] l. définitive / %s: fetching last AN hemi or last CMP commission' % curr_step.get('step'))
        for i in reversed(range(curr_step_index)):
            step = steps[i]
            if step.get('echec'):
                continue
            if step.get('stage') == 'CMP' and step.get('step') == 'commission':
                return i
            if step.get('institution') == 'assemblee' and step.get('stage') == 'nouv. lect.' and step.get('step') == 'hemicycle':
                return i
        else:
            raise Exception('[step_logic] l. définitive / depot: no good text found, this should never happen !')

    cmp_hemi_failed = False
    i = curr_step_index
    while i > 0:
        i -= 1
        step = steps[i]

        if step.get('stage') == 'CMP' and step.get('step') == 'hemicycle':
            # a CMP hemicycle previous step is the CMP commission
            if curr_step.get('stage') == 'CMP' and curr_step.get('step') == 'hemicycle':
                continue
            # if a CMP hemi fail, we mark it to ignore the commission text and the other CMP hemi
            if step.get('echec'):
                print('[step_logic] CMP hemicycle failed, we ignore the CMP commission')
                cmp_hemi_failed = True
            if cmp_hemi_failed:
                continue

        # the amendments are done on the depot text for a renvoi en commission
        if step.get('echec') == 'renvoi en commission':
            i -= 1 # we skip the commission step
            continue

        if not step.get('echec'):
            # for the old procedure, there's no text produced during the commission
            if (is_old_procedure or use_old_procedure(step)) and step.get('step') == 'commission' and step.get('stage') != 'CMP':
                continue

            # if a CMP hemi fail, we ignore the CMP commission
            if cmp_hemi_failed and step.get('stage') == 'CMP' and step.get('step') == 'commission' and curr_step.get('stage') != 'CMP':
                continue

            # do not take previous depot if it's not one of the initial depot
            if not get_depot_step and step.get('step') == 'depot' and not is_one_of_the_initial_depots(steps, i):
                continue

            return i



"""
Common law parser for AN/Sénat

Run with parse-texte <URL>
Outputs results to stdout
"""

import copy
import re
import sys

from bs4 import BeautifulSoup
from lawfactory_utils.urls import download

from .common import get_text_id, upcase_accents, real_lower
from .sort_articles import bister


class TextParsingFailedException(Exception):
    pass

# inspired by duralex/alinea_parser.py
def word_to_number(word):
    words = {
        'premiere': 1,
        'deuxieme': 2,
        'seconde': 2,
        'troisieme': 3,
        'quatrieme': 4,
        'cinquieme': 5,
        'sixieme': 6,
        'septieme': 7,
        'huitieme': 8,
        'neuvieme': 9,
        'dixieme': 10,
        'onzieme': 11,
        'douzieme': 12,
        'treizieme': 13,
        'quatorzieme': 14,
        'quinzieme': 15,
        'seizieme': 16,
    }

    for i, let in enumerate('abcdefh'):
        words[let] = i + 1

    word = real_lower(word).replace('è', 'e')
    if word in words:
        return str(words[word])


def non_recursive_find_all(node, should_be_parsed, should_be_ignored):
    """
    if there's a <p> inside a <p>, we don't want to process both
    so we stop at the first top-level <p> we find
    and ignore the children
    """
    if should_be_parsed(node):
        yield node
    elif should_be_ignored(node):
        return
    elif hasattr(node, 'children'):
        for child in node.children:
            yield from non_recursive_find_all(child, should_be_parsed, should_be_ignored)


def clean_extra_expose_des_motifs(html):
    """
    the budget related texts have an exposé des motifs per article
    at the depot step, we remove all of them except the last one
    to make the later parsing easier
    """
    before_expose, after_expose = [], []
    last_expose = []
    expose = False
    count = 0
    we_are_inside_an_expose_table = False
    for line in html.split('\n'):
        if '>Exposé des motifs' in line:
            expose = ['']
            count += 1
        # detect end of exposé
        elif line and expose:
            if '<table' in line:
                we_are_inside_an_expose_table = True
            if '</table' in line:
                we_are_inside_an_expose_table = False
            if not we_are_inside_an_expose_table:
                if '"text-align: center">' in line or \
                        '<b>' in line or \
                        '</a>' in line or \
                        '<a name=' in line:
                    last_expose = expose
                    before_expose += after_expose
                    after_expose = []
                    expose = False
        if not expose:
            after_expose.append(line)
        else:
            expose.append(line)
    if count > 3:
        return '\n'.join(before_expose + last_expose + after_expose), True
    return html, False


# Warning changing parenthesis in this regexp has multiple consequences throughout the code
section_titles = "((chap|t)itre|partie|volume|livre|tome|(sous-)?section)"

re_definitif = re.compile(r'<p([^>]*align[=:\s\-]*center"?)?>\(?<(b|strong)>\(?texte d[^f]*finitif\)?</(b|strong)>\)?</p>', re.I)
re_definitif_new_format = re.compile(r'<span [^>]*font-weight: bold;[^>]*>\(Texte d[^f]*finitif\)</span>', re.I) # embedded HTML from AN /textes/
definitif_before_congres = "<i>(Texte voté par les deux Assemblées du Parlement en termes identiques ; ce projet ne deviendra définitif, conformément à l'article 89 de la Constitution, qu'après avoir été approuvé par référendum ou par le Parlement réuni en Congrès)</i>"
definitif_after_congres = "Le Parlement, réuni en Congrès, a approuvé dans les conditions prévues à l'article 89, alinéa 3, de la Constitution, le projet de loi constitutionnelle dont la teneur suit"

clean_texte_regexps = [
    (re.compile(r'[\n\t\r\s]+'), ' '),
    # (re.compile(r'(<t[rdh][^>]*>) ?<p [^>]*> ?'), r'\1'), # warning: this was to clean tables but the
    # (re.compile(r' ?</p> ?(</t[rdh]>)'), r'\1'),          #          conclusion of report can be in a table too
    (re.compile(r'(>%s\s*[\dIVXLCDM]+(<sup>[eE][rR]?</sup>)?)\s+-\s+([^<]*?)\s*</p>' % section_titles.upper()), r'\1</p><p><b>\6</b></p>'),
    (re.compile(r'(<sup>[eE][rR]?</sup>)(\w+)'), r'\1 \2'), # add missing space, ex: "1<sup>er</sup>A "
    (re.compile(r'(\w)<br/?>(\w)'),  r'\1 \2'), # a <br/> should be transformed as a ' ' only if there's text around it (visual break)
    (re.compile(r'<(em|s)> </(em|s)>'),  r' '), # remove empty tags with only one space inside
    (re.compile(r'(<p[^>]*><(b|strong)>Article[^<]*</(b|strong)></p>) \1'), r'\1'), # duplicate article title i.e. https://www.senat.fr/leg/tas10-156.html art 26
    (re.compile(r'(<a name=[\'"])[^\'"]+([\'"]>)', re.I), r'\1\2'),        # we use '<a name=' to recognize titles but we never use the anchor value so we can remove it to handle the following duplicates
    (re.compile(r'((<p style="text-align: center">(<[^>]+>)*[^<]+(<[^>]+>)*</p> ){2,})\1'), r'\1'), # duplicate group of title lines i.e. http://www.assemblee-nationale.fr/14/ta-commission/r3909-a0.asp art 10 A
    (re.compile(r'((?:<(?:p|span)[^>]*>\s*)+[^<]+)(?:<(?:b|strong|br)>\s*)+(Article \d[^<]{0,10})(?:</(?:b|strong)>)?((?:</(?:p|span)>)+)', re.I), r'\1\3<p><b>\2</b></p>'), # article title in previous article text i.e. http://www.assemblee-nationale.fr/13/ta/ta0173.asp art 9
    (re.compile(r"<p data-pastille=.*?</p>"), ""), # remove pastilles coming from AN "/textes/" HTML sometimes copy-pasted into Senate pages
]

re_clean_title_legif = re.compile(r"[\s|]*l[eé]gifrance(.gouv.fr)?$", re.I)
clean_legifrance_regexps = [
    (re.compile(r'[\n\t\r\s]+'), ' '),
    (re.compile(r'<a[^>]*>\s*En savoir plus sur ce[^<]*</a>', re.I), ''),
    (re.compile(r'<a/?[^>]*>', re.I), ''),
    (re.compile(r'\s*<br/>\s*', re.I), '</p><p>'),
    (re.compile(r'<div[^>]*class="titreSection[^>]*>\s*(%s\s+[\dIVXLCDM]+e?r?)\s*:\s*([^<]*?)\s*</div>' % section_titles, re.I), r'<p>\1</p><p><b>\5</b></p>'),
    (re.compile(r'<div[^>]*class="titre(?:Art|Section)[^>]*>([^<]*?)\s*</div>', re.I), r'<p><b>\1</b></p>'),
    (re.compile(r'\[Dispositions (déclarées non conformes à la Constitution( à compter du .{0,30})? par|résultant de) la (<a[^>]*?>)?décision( du Conseil constitutionnel)? n° \d+-\d+ ?(<\/a>)? ?DC du .{0,30}\]', re.I), "(Censuré)"),
    (re.compile(r'\[Rédaction conforme .{0,100} la (<a[^>]*?>)?décision( du Conseil constitutionnel)? n° \d+-\d+ ?(<\/a>)? ?DC du .{0,30}\]\.?', re.I), ""),
    (re.compile(r'―'), '-'),
    (re.compile(r'([\.\s]+)-([^\s\-]+)'), r'\1 - \2'),
]


# Convert from roman numbers
re_mat_romans = re.compile(r"[IVXCLDM]+", re.I)
romans_map = list(zip(
    (1000,  900, 500, 400 , 100,  90 , 50 ,  40 , 10 ,   9 ,  5 ,  4  ,  1),
    ( 'M', 'CM', 'D', 'CD', 'C', 'XC', 'L', 'XL', 'X', 'IX', 'V', 'IV', 'I')
))


def romans(n):
    n = n.upper()
    i = res = 0
    for d, r in romans_map:
        while n[i:i + len(r)] == r:
            res += d
            i += len(r)
    return res


def lower_but_first(text):
    return text[0].upper() + real_lower(text[1:])


re_fullupcase = re.compile(r"^([\W0-9]*)([A-Z%s][\W0-9A-Z%s]*)$" % (upcase_accents, upcase_accents), re.U)


def clean_full_upcase(text):
    mat = re_fullupcase.match(text)
    if mat:
        text = mat.group(1) + lower_but_first(mat.group(2))
    return text

re_clean_premier = re.compile(r'((PREM)?)(1|I)ER?')
re_clean_bister = re.compile(r'([IXV\d]+e?r?)\s+(%s)' % bister, re.I)
re_clean_subsec_space = re.compile(r'^("?[IVX0-9]{1,4}(\s+[a-z]+)?(\s+[A-Z]{1,4})?)\s*([\.°\-]+)\s*([^\s\)])', re.I)
re_clean_subsec_space2 = re.compile(r'^("?[IVX0-9]{1,4})\s*([a-z]*)\s*([A-H]{1,4})([\.°\-])', re.I)
re_clean_punc_space = re.compile(r'([°«»:;,\.!\?\]\)%€&\$])([^\s\)\.,\d"])')
re_clean_spaces = re.compile(r'\s+')
re_clean_coord = re.compile(r'^(<i>)?([\["\(\s]+|pour)*coordination[\]\)\s\.]*(</i>)?', re.I)
re_préliminaire = re.compile(r' pr..?liminaire', re.I)
# Clean html and special chars
lower_inner_title = lambda x: x.group(1)+lower_but_first(x.group(3))+" "
html_replace = [
    (re.compile(r"−"), "-"),
    (re.compile(r" "), " "),
    (re.compile(r"<!--.*?-->", re.I), ""),
    (re.compile(r"<span[^>]*color: #(0070b9|006fb9)[^>]*>\(\d+\)\s*</span>", re.I), ""), # remove pastilles
    (re.compile(r"<span[^>]*color: white[^>]*>.*?</span>", re.I), ""), # remove invisible text
    (re.compile(r"(<img[^>]*>\s*<br/>\s*)", re.I), ""), # remove <img><br/> before the next regex kills my precious '«'
    (re.compile(r"</?br/?>\s+", re.I), " "),
    (re.compile(r'(«\s+|\s+»)'), '"'),
    (re.compile(r'(«|»|“|”|„|‟|❝|❞|＂|〟|〞|〝)'), '"'),
    (re.compile(r"(’|＇|’|ߴ|՚|ʼ|❛|❜)"), "'"),
    (re.compile(r"(‒|–|—|―|⁓|‑|‐|⁃|⏤)"), "-"),
    (re.compile(r"(</?\w+)[^>]*>"), r"\1>"), # removes html attributes
    (re.compile(r"(</?)em>", re.I), r"\1i>"),
    (re.compile(r"(</?)strong>", re.I), r"\1b>"),
    (re.compile(r"<(![^>]*|/?(p|span))>", re.I), ""),
    (re.compile(r"\s*\n+\s*"), " "),
    (re.compile(r"<[^/>]*></[^>]*>"), ""),
    (re.compile(r"^<b><i>", re.I), "<i><b>"),
    (re.compile(r"</b>(\s*)<b>", re.I), r"\1"),
    (re.compile(r"<a>(\s*)</a>", re.I), r"\1"), # remove bad copy pastes of <a name=..> </a> where the content has been removed
    (re.compile(r"</?sup>", re.I), ""),
    (re.compile(r"^((<[bi]>)*)\((S|AN)[12]\)\s*", re.I), r"\1"),
    (re.compile(r"<s>(.*)</s>", re.I), ""),
    (re.compile(r"^(<b>Article\s*)\d+\s*(?:<s>\s*)+", re.I), r"\1"), # take old barred number of an article instead of renumbered one
    (re.compile(r"</?s>", re.I), ""),
    (re.compile(r"\s*</?img>\s*", re.I), ""),
    (re.compile(r"œ([A-Z])"), r"OE\1"),
    (re.compile(r"œ\s*", re.I), "oe"),
    (re.compile(r'^((<[^>]*>)*")%s ' % section_titles, re.I), lower_inner_title),
    (re_préliminaire, ' préliminaire'),
    (re.compile(r'<strike>[^<]*</strike>', re.I), ''),
    (re.compile(r'^<a>(\w)', re.I), r"\1"),
    (re.compile(r'^[.…\s]+(((suppr|conforme)[^\.…]{0,10})+)[.…\s]+$', re.I), r"\1"),  # clean "......Conforme....." to "Conforme"
    (re.compile(r'(\w\s*(?:\</[^>]*>)*\s*)\.{10,}(\s*;)?(</i>)?$', re.I), r"\1\2\3"),  # clean "III. - <i>Conform[e</i>.......]" to "III. - <i>Conform[e</i>]"
    (re_clean_spaces, " ")
]


def clean_html(t):
    for i, (regex, repl) in enumerate(html_replace):
        try:
            t = regex.sub(repl, t)
        except Exception as e:
            raise TextParsingFailedException("Crashed while applying regexp", regex, "with replacement", repl, "to", t, e)
    return t.strip()

re_clean_et = re.compile(r'(,|\s+et)\s+', re.I)

def check_section_is_not_a_duplicate(section_id, articles):
    for block in articles:
        assert not (block['type'] == 'section' and block.get('id') == section_id)

re_move_table_guillemets_left = re.compile(r'^(<table[^>]*>(?:<thead[^>]*>.*?</thead>)?(?:<tbody[^>]*>)?<tr[^>]*>)<td[^>]*>\s*"\s*</td>', re.I)
re_move_table_guillemets_right = re.compile(r'<td[^>]*>\s*("\.?)\s*</td>(</tr>(?:</tbody>)?</table>)$', re.I)
re_move_table_guillemets_within = re.compile(r'^(<table[^>]*>(?:<thead[^>]*>.*?</thead>)?(?:<tbody[^>]*>)?<tr[^>]*><td[^>]*>)\s*"\s*([^"]+)\s*("\.?)\s*(</td></tr>(?:</tbody>)?</table>)$', re.I)

def add_to_articles(dic, all_articles):
    # Clean empty articles with only "Supprimé" as text
    if not dic:
        return
    if 'alineas' in dic:
        # check for duplicates
        for article in all_articles:
            if dic.get('titre') and dic.get('titre') == article.get('titre') and 'source_text' not in article:
                raise TextParsingFailedException('Duplicate article title found: %s', article.get('titre'))

        if len(dic['alineas']) == 1 and dic['alineas']['001'].startswith("(Supprimé)"):
            dic['statut'] = "supprimé"
            dic['alineas'] = {'001': ''}
        elif len(dic['alineas']) == 1 and dic['alineas']['001'].startswith("(Censuré)"):
            dic['statut'] = "supprimé"
            dic['alineas'] = {'001': '(Censuré)'}
        elif dic['statut'].startswith('conforme') and not len(dic['alineas']):
            dic['alineas'] = {'001': '(Non modifié)'}
        # Assume an article is non-modifié if it's empty (but check if it's supprimé)
        # but there's a know side-effect, it may generate non-modifié articles of deleted
        # articles like in the text for article 35 bis:
        #   https://www.senat.fr/rap/l09-567/l09-5671.html
        # or "non-modifié" during the depot for an empty article:
        #   https://www.senat.fr/leg/pjl14-661.html - article 53
        elif not dic['statut'].startswith('suppr') and not len(dic['alineas']):
            dic['alineas'] = {'001': '(Non modifié)'}
        multiples = re_clean_et.sub(',', dic['titre']).split(',')
        if len(multiples) > 1:
            for d in multiples:
                new = copy.deepcopy(dic)
                new['titre'] = d
                all_articles.append(new)
            return
        # Cleanup guillemets around tables
        if '<table' in "".join(dic["alineas"].values()):
            als = {}
            i = 1
            prevguil = None
            prevtabl = None
            for aln in sorted(dic['alineas'].keys()):
                al = dic['alineas'][aln].strip()
                if al in ['"', '".', '";']:
                    if prevtabl:
                        if '"' not in als['%03d' % (i-1)][-2:]:
                            als['%03d' % (i-1)] += al
                        prevtabl = None
                        continue
                    prevguil = al
                    continue
                if '<table' in al:
                    al = re_move_table_guillemets_left.sub(r'"\1', al)
                    al = re_move_table_guillemets_right.sub(r'\2\1', al)
                    al = re_move_table_guillemets_within.sub(r'"\1\2\4\3', al)
                    prevtabl = True
                    if prevguil and not al.startswith('"'):
                        al = '"' + al
                else:
                    prevtabl = None
                prevguil = False
                als['%03d' % i] = al
                i += 1
            dic['alineas'] = als
    all_articles.append(copy.deepcopy(dic))


blank_none = lambda x: x if x else ""
re_cl_html = re.compile(r"<[^>]+>")
re_cl_html_except_tables = re.compile(r"</?[^t/][^>]*>", re.I)
re_fix_missing_table = re.compile(r'(<td>\W*)$', re.I)
cl_html_except_tables = lambda x: re_fix_missing_table.sub(r'\1</td></tr></tbody></table>', re_cl_html_except_tables.sub('', x)).strip().replace('> ', '>').replace(' <', '<').replace('<td><tr>', '<td></td></tr><tr>')
re_cl_par  = re.compile(r"[()\[\]]")
re_cl_uno  = re.compile(r"(premie?r?|unique?)", re.I)
re_cl_sec_uno = re.compile(r"^[Ii1][eE][rR]?")
re_cl_uno_uno = re.compile(r"^1(\s|$)")
re_mat_sec = re.compile(r"(?:<b>)?%s(\s+([^:-]+)e?r?)(?:[:-]\s+(?P<titre>[^<]*))?(?:</b>)?" % section_titles, re.I)
re_cl_sec_simple_num = re.compile(r"(?:<(?P<tag>i|b)>)?(?P<num>[A-Z]|[IVX]{,5})\. - (?P<titre>[^<]+)", re.I) # section name like "B. - XXX" or "<i>IV. - XXX"
re_cl_sec_part = re.compile(r"^(?:<b>)?(?P<num>\w{,11})\s+partie\s*(?::(?P<titre>[^<]*))?(?:</b>)?$", re.I) # partie name like "cinquiéme partie : XXXX"
re_mat_n = re.compile(r"((pr..?)?limin|unique|premier|[IVX\d]+)", re.I)
re_mat_art = re.compile(r"articles?(?!\sL\.)\s*([^(]*)(\([^)]*\))?$", re.I)
re_mat_ppl = re.compile(r"((<(b|h[12])>)?\s*pro.* (loi|résolution)|<h2>\s*pro.* (loi|résolution)\s*</\3>)", re.I)
re_mat_tco = re.compile(r"\s*<(b|h[12])>\s*(ANNEXE[^:]*:\s*|\d+\)\s+|[IVX]+\.\s+|<a name[^>]*>\s*</a>\s*)*TEXTES?\s*(([ÉE]LABOR|ADOPT)[EÉ]S?\s*PAR|DE)\s*LA\s*COMMISSION.*(</\1>\s*$|\(.*\))")
re_mat_exp = re.compile(r"(<(b|strong)>)?\s*(expos[eéÉ]|table des matières)", re.I)
re_mat_end = re.compile(r"((<i>)?Délibéré en|(<i>)?NB[\s:<]+|(<b>)?RAPPORT ANNEX|(<b>)?États législatifs annexés|(<(i|t\w+)>\s*)*Fait à .*, le|\s*©|\s*N.?B.?\s*:|(</?i>)*<a>[1*]</a>\s*(</?i>)*\(\)(</?i>)*|<i>\(1\)\s*Nota[\s:]+|La présente loi sera exécutée comme loi de l'Etat|<a>\*</a>\s*(<i>)?1)", re.I)
re_mat_ann = re.compile(r"\s*<b>\s*ANNEXES?[\s<]+", re.I)
re_mat_dots = re.compile(r"^(<i>)?([.…_]\s?)+(</i>)?$")
re_mat_st = re.compile(r"(<i>\s?|\(|\[)+(texte)?\s*(conform|non[\s\-]*modif|suppr|nouveau).{0,30}$", re.I)
re_mat_new = re.compile(r"\s*\(\s*nouveau\s*\)\s*", re.I)
re_mat_texte = re.compile(r'\((Adoption du )?texte (modifié|élaboré|voté|d(u|e l))', re.I)
re_approb = re.compile(r"^Est autorisée l'approbation de l", re.I)
re_mat_single_char = re.compile(r'^\s*[LMN]\s*$')
re_clean_idx_spaces = re.compile(r'^([IVXLCDM0-9]+)\s*\.\s*')
re_clean_art_spaces = re.compile(r'^\s*("?)\s+')
re_clean_art_spaces2 = re.compile(r'\s+\.\s*-\s+')
re_clean_conf = re.compile(r"(?:\(|^([IVX]{1,3}(?: et [IVX]{1,3})?\. - )?)(conforme|non[\s-]*modifi..?)s?(?:\)|$)", re.I)
re_clean_supr = re.compile(r'(?:\(|^)(dispositions?\s*d..?clar..?es?\s*irrecevable.*article 4.*Constitution.*|(maintien de la |Article )?suppr(ession|im..?s?)(\s*(conforme|maintenue|par la commission mixte paritaire))*)\)?[\"\s]*$', re.I)
re_echec_hemi = re.compile(r"L('Assemblée nationale|e Sénat) (a rejeté|n'a pas adopté)[, ]+", re.I)
re_echec_hemi2 = re.compile(r"de loi (a été rejetée?|n'a pas été adoptée?) par l('Assemblée nationale|e Sénat)\.$", re.I)
re_echec_hemi3 = re.compile(r"le Sénat décide qu'il n'y a pas lieu de poursuivre la délibération", re.I)
re_echec_com = re.compile(r"(la commission|elle) .*(effet est d'entraîner le rejet|demande de rejeter|a rejeté|n'a pas adopté|n'a pas élaboré|décide de ne pas présenter de conclusions? sur|rejette l'ensemble|ne pas établir|ne pas adopter)[dleau\s]*(projet|proposition|texte)[.\s]", re.I)
re_echec_com2 = re.compile(r"L'ensemble de la proposition de loi est rejeté dans la rédaction issue des travaux de la commission.", re.I)
re_echec_com3 = re.compile(r"la commission (a décidé de déposer une|adopte la) motion tendant à opposer la question préalable", re.I)
re_echec_com4 = re.compile(r"l(a|es) motions?( | .{0,5} )tendant à opposer la question préalable (est|sont) adoptées?", re.I)
re_echec_com5 = re.compile(r"(la|votre) commission (n'a pas élaboré|a décidé de ne pas adopter) [dleau\s]*(projet|proposition|texte)", re.I)
re_echec_com6 = re.compile(r"[dleau\s]*(projet|proposition|texte) est (considéré comme )?rejetée? par la commission", re.I)
re_echec_cmp = re.compile(r" (a conclu à l'échec de ses travaux|(ne|pas) .*parven(u[es]?|ir) à (élaborer )?un texte commun)", re.I)
re_rap_mult = re.compile(r'[\s<>/ai]*N[°\s]*\d+\s*(,|et)\s*[N°\s]*\d+', re.I)
re_src_mult = re.compile(r'^[\-\s]*L(?:A PROPOSITION|E PROJET) DE LOI (?:ADOPTÉ PAR LE SÉNAT )?[Nn]°\s*(\d+)\D')
re_clean_mult_1 = re.compile(r'\s*et\s*', re.I)
re_clean_mult_2 = re.compile(r'[^,\d]', re.I)
re_clean_footer_notes = re.compile(r"[\.\s]*\(*\d*\([\d\*]+[\)\d\*\.\s]*$")
re_sep_text = re.compile(r'\s*<b>\s*(article|%s)\s*(I|uniqu|pr..?limina|1|prem)[ier]*\s*</b>\s*$' % section_titles, re.I)
re_stars = re.compile(r'^[\s*_]+$')
re_art_uni = re.compile(r'\s*article\s*unique\s*$', re.I)


def normalize_1(name, one):
    name = name.strip()
    name = re_cl_uno.sub(one, name)
    name = re_cl_sec_uno.sub(one, name)
    name = re_cl_uno_uno.sub(one + r'\1', name)
    return name.strip().strip(" -'")


def normalize_section_title(line, line_soup, has_multiple_expose):
    # transforms "Xeme partie (: <titre>)" to "partie Xeme (: <titre>)"
    m = re_cl_sec_part.match(line)
    if m:
        line = 'partie ' + m.group('num')
        if m.group('titre'):
            line += ' : ' + m.group('titre')
        return line
    # reformats A, B, C, I, II, III sections
    m = re_cl_sec_simple_num.match(line)
    if m:
        # those sections either start with <b> or <i>
        # hack: a "<a name=XXX>" in the html is sufficient
        #       only for depot for pjlf/plfss (multiple expose)
        #       since there are bugs in the AN HTML
        #       ex: alinea III. in article 11 having <a name="XXX">:
        #       http://www.assemblee-nationale.fr/14/ta/ta0208.asp
        if m.group('tag') or ('<a name=' in str(line_soup) and has_multiple_expose):
            # treats A, B, C as sub-sections and I, II, III as sections
            type = 'sous-section' \
                if m.group('tag') == 'b' or re.match(r'[A-H]', m.group('num')) \
                else 'section'
            return '%s %s : %s' % (type, m.group('num'), m.group('titre'))
    return line


def clean_article_name(text):
    # Only keep first line for article name
    # but to do that while keeping the regexes the same
    # we need to add our own marker
    NEW_LINE_MARKER = 'NEW_LINE_MARKER'
    html = str(text)
    html = re.sub(r'<br/?>', NEW_LINE_MARKER, html)
    line = clean_html(html)
    cl_line = re_cl_html.sub("", line).strip()
    cl_line = [l for l in cl_line.split(NEW_LINE_MARKER) if l.strip()][0]

    # If there's a ':', what comes after is not related to the name
    cl_line = cl_line.split(':')[0].strip()

    # simple cleaning of 1er. -> 1er
    cl_line = cl_line.rstrip('.')

    return cl_line

def parse(url, resp=None, DEBUG=False, include_annexes=False):
    """
    parse the text of an url, an already cached  to`resp` can be passed to avoid an extra network request
    """
    all_articles = []
    def pr_js(article):
        nonlocal all_articles, texte
        if not len(all_articles):
            add_to_articles(texte, all_articles)
        add_to_articles(article, all_articles)

    if url.endswith('.pdf'):
        print("WARNING: text url is a pdf: %s skipping it..." % url)
        return all_articles
    if 'assemblee-nat.fr' in url:
        print("WARNING: url corresponds to old AN website: %s skipping it..." % url)
        return all_articles


    if url.startswith('http'):
        resp = download(url) if resp is None else resp
        if '/textes/'in url:
            resp.encoding = 'utf-8'
        if 'assemblee-nationale.fr' in url:
            resp.encoding = 'Windows-1252'
        string = resp.text
    elif url == '-':
        string = sys.stdin.read()
    else:
        try:
            string = open(url).read()
        except:
            string = open(url, encoding="Windows-1252").read()

    string, has_multiple_expose = clean_extra_expose_des_motifs(string)

    if 'legifrance.gouv.fr' in url:
        for reg, res in clean_legifrance_regexps:
            string = reg.sub(res, string)
    else:
        for reg, res in clean_texte_regexps:
            string = reg.sub(res, string)

    #fix weird Sénat formatting with single cells tables around pieces of text sometimes multiline... ex: https://www.senat.fr/leg/ppl15-246.html
    for match in re.findall(r'(<table[^>]*>\s*(?:<t(?:body|r|d|h)[^>]*>\s*)+)(.*?)((?:\s*</t(?:body|r|d|h)[^>]*>)+\s*</table>)', string, re.I):
        if not re.search(r'<t(r|d|h)[^>]*>', match[1], re.I):
            string = string.replace(''.join(match), match[1])

    srclst = []
    source_avenants = False
    m = re.search(r"NB(\s|&nbsp;)+:(\s|&nbsp;)+[lL]es? textes? d(u |es |e la |e l&#8217;)((convention|traité|avenant)s? et de(s| l&#8217;))?(accord|convention)s?(-cadres?)? figuren?t? (respectivement )?en annexe aux (deux |trois )?projets de loi \(n", re.sub(r'</?span[^>]*>', '', string), re.I)
    if m:
        try:
            srclst = [int(s.strip('no ')) for s in (
                    string.replace('<sup>', '').replace('</sup>', '').replace('&nbsp;', ' ')
                    .replace('aux deux projets', 'aux projets').replace('aux trois projets', 'aux projets')
                    .replace('°', 'o').replace('nos ', 'no ').replace('ns ', 'no ').replace('(n ', '(no ')
                    .split(' en annexe aux projets de loi (no ')[1]
                    .strip()
                    .split(')')[0]
                    .strip()
                    .replace(' et ', ', ')
                    .split(', '))]
            source_avenants = True
        except Exception as e:
            if DEBUG:
                print("WARNING, multi-reports detected with NB method crashing (%s: %s), trying regular method..." % (type(e), e))
    if not source_avenants and "/rapports/r" in url and "TEXTES ADOPTÉS PAR LA COMMISSION" in string and string.count(">Article unique<") == 2:
        m = re.search(r'<i>Assemblée nationale&nbsp;:&nbsp;</i><b>(\d+) </b>et<b> (\d+)</b>', string)
        if m:
            srclst = [int(m.group(1)), int(m.group(2))]
            source_avenants = True

    definitif = re_definitif.search(string) is not None or re_definitif_new_format.search(string) is not None or 'legifrance.gouv.fr' in url
    soup = BeautifulSoup(string, "html5lib")
    texte = {"type": "texte", "source": url, "definitif": definitif}

    # Generate Senat or AN ID from URL
    if url.startswith('http'):
        if "legifrance.gouv.fr" in url:
            m = re.search(r"cidTexte=(JORFTEXT\d+)(\D|$)", url, re.I)
            if m:
                texte["id"] = m.group(1)
            elif "/jo/texte" in url:
                texte["id"] = url.split('/')[-3]
        elif re.search(r"assemblee-?nationale", url, re.I):
            m = re.search(r"/(\d+)/.+/(ta)?[\w\-]*(\d{4})[\.\-]", url, re.I)
            numero = int(m.group(3))
            texte["id"] = "A" + m.group(1) + "-"
            if m.group(2) is not None:
                texte["id"] += m.group(2)
            texte["id"] += str(numero)
            texte["nosdeputes_id"] = get_text_id(url)
        else:
            m = re.search(r"(ta|l)?s?(\d\d)-(\d{1,3})(rec)?\d?(_mono)?\.", url, re.I)
            if m is None:
                m = re.search(r"/(-)?20(\d+)-\d+/(\d+)(_mono)?.html", url, re.I)
            numero = int(m.group(3))
            texte["id"] = "S" + m.group(2) + "-"
            if m.group(1) is not None:
                texte["id"] += m.group(1)
            texte["id"] += "%03d" % numero
            texte["nossenateurs_id"] = get_text_id(url)

    texte["titre"] = clean_html(re_clean_title_legif.sub('', soup.title.string.strip())) if soup.title else ""
    texte["expose"] = ""
    expose = False

    # states 'read' can be set to:
    READ_DISABLED = -1 # the text is not detected yet
    READ_TEXT = 0 # read the text
    READ_TITLE = 1 # titles lecture
    READ_ALINEAS = 2 # alineas lecture

    read = READ_TEXT
    art_num = ali_num = 0
    article = {}
    indextext = -1
    curtext = -1
    section = {"type": "section", "id": ""}

    rejected_all_articles = [] # we only keep the last detected text by default, here are stored the previous texts

    def should_be_parsed(x):
        """returns True if x can contain useful information"""
        if x.name not in ('p', 'table', 'h1', 'h2', 'h4'):
            return False
        # hack: we don't want to parse the table containing the conclusion from the senat
        #       ex: https://www.senat.fr/leg/tas12-040.html
        if x.name == "table" and re.search("SESSION (EXTRA)?ORDINAIRE DE", str(x)):
            return False
        # hack: senate can copy paste the /textes/ output from the AN
        #       ex: https://www.senat.fr/leg/ppl17-545.html
        # TODO: they also mess up the encoding by doing that
        if x.name == "table" and re.search("<!-- Col de controle de taille -->", str(x)):
            return False
        return True


    def should_be_ignored(x):
        if hasattr(x, 'attrs') and 'display: none' in x.attrs.get('style', ''):
            return True
        return False


    for text in non_recursive_find_all(soup, should_be_parsed, should_be_ignored):
        line = clean_html(str(text))
        if DEBUG:
            print(read, article.get('titre') or art_num, ali_num, line, file=sys.stderr)

        # limit h2/h4 matches to PPL headers or Article unique
        if text.name not in ('p', 'table') and not re_mat_ppl.match(line) and not re_mat_tco.match(line) and 'Article unique' not in line:
            if DEBUG:
                print(" -> IGNORING LINE", file=sys.stderr)
            continue

        if re_stars.match(line):
            continue
        if line == "<b>RAPPORT</b>" or line == "Mesdames, Messieurs,":
            read = READ_DISABLED
        if (srclst or indextext != -1) and re_sep_text.match(line):
            curtext += 1
            art_num = 0
        srcl = re_src_mult.search(line)
        if not source_avenants and srcl and read in (READ_DISABLED, READ_TEXT):
            srclst.append(int(srcl.group(1)))
            continue
        cl_line = re_cl_html.sub("", line).strip()
        if re_rap_mult.match(line):
            line = cl_line
            line = re_clean_mult_1.sub(",", line)
            line = re_clean_mult_2.sub("", line)
            cl_line = re_cl_html.sub("", line).strip()
            for n_t in line.split(','):
                indextext += 1
                if int(n_t) == numero:
                    break
        elif re_mat_ppl.match(line) or re_mat_tco.match(line) or (
                read == READ_DISABLED and line == "<b>Article 1er</b>"):
            read = READ_TEXT
            if len(all_articles):
                if DEBUG:
                    print('WARNING: Found articles before the real text')
                if article is not None:
                    pr_js(article)
                rejected_all_articles.append(all_articles)
                all_articles = []
                article = {}
                art_num = 0
        elif re_mat_exp.match(line):
            read = READ_DISABLED # Deactivate description lecture
            expose = True
        elif read == READ_TEXT and definitif_before_congres in line or definitif_after_congres in line:
            texte['definitif'] = True
            if all_articles:
                all_articles[0]['definitif'] = True
            continue
        elif (re_echec_cmp.search(cl_line)
                or re_echec_com.search(cl_line)
                or re_echec_com2.search(cl_line)
                or re_echec_com3.search(cl_line)
                or re_echec_com4.search(cl_line)
                or re_echec_com5.search(cl_line)
                or re_echec_com6.search(cl_line)
                or re_echec_hemi.match(cl_line)
                or re_echec_hemi2.search(cl_line)
                or re_echec_hemi3.search(cl_line)
            ) and 'dont la teneur suit' not in cl_line:
            pr_js({"type": "echec", "texte": cl_line})
            break
        elif read == READ_DISABLED:
            continue
        # or (indextext != -1 and curtext != indextext): #keep all texts resulting from multireport now it's selected then in complete

        # crazy edge case: "(Conforme)Article 24 bis A (nouveau)" on one line
        # http://www.assemblee-nationale.fr/13/projets/pl3324.asp
        # simplified, just do the "(Conforme)" case
        if '<i>(Conforme)</i>' in line and re_mat_art.search(line):
            article["statut"] = 'conforme'
            line = line.replace('<i>(Conforme)</i>', '')
            cl_line = cl_line.replace('(Conforme)', '')

        # another crazy edge case: the text is inside the annexe
        # ex: http://www.assemblee-nationale.fr/13/rapports/r2083.asp
        # TODO: could detect via "le présent projet de loi dans le texte figurant en annexe"
        #       like the source_avenants logic
        if read != READ_ALINEAS and re_approb.match(line):
            art_num += 1
            article = {
                "type": "article",
                "order": art_num,
                "alineas": {},
                "statut": "none",
                "titre": "1er"
            }
            read = READ_ALINEAS

        # Identify section zones
        line = normalize_section_title(line, text, has_multiple_expose)
        m = re_mat_sec.match(line)
        if m:
            read = READ_TITLE # Activate titles lecture
            section["type_section"] = real_lower(m.group(1))
            section_typ = m.group(1).upper()[0]
            if m.group(3) is not None:
                section_typ += "S"

            if re.search(re_préliminaire, line) or " LIMINAIRE" in line.upper():
                section_num = "L"
            else:
                section_num = re_cl_html.sub('', m.group(5).strip())
                if word_to_number(section_num) is not None:
                    section_num = word_to_number(section_num)
                section_num = normalize_1(section_num, '1')
                section_num = re_clean_bister.sub(lambda m: m.group(1)+" "+real_lower(m.group(2)), section_num)
                section_num = re_mat_new.sub('', section_num).strip()
                m2 = re_mat_romans.match(section_num)
                if m2:
                    rest = section_num.replace(m2.group(0), '')
                    section_num = romans(m2.group(0))
                    if rest:
                        section_num = str(section_num) + rest
            # Get parent section id to build current section id
            section_par = re.sub(r"" + section_typ + r"[\dL].*$", "", section["id"])
            section["id"] = section_par + section_typ + str(section_num)
            # check_section_is_not_a_duplicate(section["id"])

            titre = blank_none(m.group('titre')).strip()
            if titre:
                section['titre'] = titre
                if article is not None:
                    pr_js(article)
                    article = {}
                pr_js(section)
                read = READ_TEXT
        elif re_mat_end.match(line) and not include_annexes:
            if not expose:
                if DEBUG:
                    print("DEBUG: END OF TEXT OF DETECTED")
                if len(all_articles) > 0:
                    break
            expose = False
            continue
        # Annexes.
        elif read == READ_ALINEAS and re_mat_ann.match(line):
            if include_annexes:
                if article is not None:
                    pr_js(article)
                titre = re_cl_html.sub("", re_mat_ann.sub("", line))
                art_num += 1
                article = {
                    "type": "annexe",
                    "order": art_num,
                    "alineas": {},
                    "statut": "none",
                    "titre": titre
                }
                ali_num = 0
            else:
                break
        # Identify titles and new article zones
        elif (re.match(r"(<i>)?<b>", line) or
                re_art_uni.match(cl_line) or
                re.match(r"^Articles? ", line)
              ) and not re.search(r">Articles? supprimé", line):

            line = cl_line.strip()
            # Read a new article
            if re_mat_art.match(line):
                if article is not None:
                    pr_js(article)
                read = READ_ALINEAS # Activate alineas lecture
                expose = False
                art_num += 1
                ali_num = 0
                article = {"type": "article", "order": art_num, "alineas": {}, "statut": "none"}
                if srclst:
                    article["source_text"] = srclst[curtext]
                m = re_mat_art.match(clean_article_name(text))
                article["titre"] = normalize_1(m.group(1), "1er").replace(u"İ", "I")

                assert article["titre"]  # avoid empty titles
                assert not texte['definitif'] or ' bis' not in article["titre"]  # detect invalid article names

                if m.group(2) is not None:
                    article["statut"] = re_cl_par.sub("", real_lower(m.group(2))).strip()
                if section["id"] != "":
                    article["section"] = section["id"]
            # Read a section's title
            elif read == READ_TITLE and line:
                section["titre"] = lower_but_first(line)
                if article is not None:
                    pr_js(article)
                    article = {}
                pr_js(section)
                read = READ_TEXT

        # detect dots, used as hints for later completion
        if read != READ_DISABLED:
            if re_mat_dots.match(line):
                if article is not None:
                    pr_js(article)
                    article = {}
                pr_js({"type": "dots"})
                read = READ_TEXT # ignore alineas after the dots
                continue

        # Read articles with alineas
        if read == READ_ALINEAS and not m:
            line = re_clean_coord.sub('', line)
            # if the line was only "Pour coordination", ignore it
            if not line:
                continue
            # Find extra status information
            if ali_num == 0 and re_mat_st.match(line):
                article["statut"] = re_cl_html.sub("", re_cl_par.sub("", real_lower(line)).strip()).strip()
                continue
            if "<table>" in line:
                cl_line = cl_html_except_tables(line)
            line = re_clean_art_spaces2.sub('. - ', re_clean_art_spaces.sub(r'\1', re_clean_idx_spaces.sub(r'\1. ', re_mat_new.sub(" ", cl_line).strip())))
            # Clean low/upcase issues with BIS TER etc.
            line = line.replace("oeUVRE", "OEUVRE")
            line = clean_full_upcase(line)
            line = re_clean_premier.sub(lambda m: (real_lower(m.group(0)) if m.group(1) else "")+m.group(3)+"er", line)
            line = re_clean_bister.sub(lambda m: m.group(1)+" "+real_lower(m.group(2)), line)
            # Clean different versions of same comment.
            line = re_clean_supr.sub('(Supprimé)', line)
            line = re_clean_conf.sub(r'\1(Non modifié)', line)
            line = re_clean_subsec_space.sub(r'\1\4 \5', line)
            line = re_clean_subsec_space2.sub(r'\1 \2 \3\4', line)

            tmp = line
            line = re_clean_punc_space.sub(r'\1 \2', tmp)
            line = re_clean_spaces.sub(' ', line)
            line = re_mat_sec.sub(lambda x: lower_but_first(x.group(1))+x.group(4) if re_mat_n.match(x.group(4)) else x.group(0), line)
            line = re_clean_footer_notes.sub(".", line)
            # Clean comments (Texte du Sénat), (Texte de la Commission), ...
            if ali_num == 0 and re_mat_texte.match(line):
                continue
            line = re_mat_single_char.sub("", line)
            line = line.strip()
            if line:
                ali_num += 1
                article["alineas"]["%03d" % ali_num] = line
        else:
            #metas
            continue

    # sometimes we find multiple text starts inside one (mainly due to annotations),
    # by default we normally keep only the latest one, but if it is empty,
    # try to find a good one from previously parsed articles
    # ex: http://www.assemblee-nationale.fr/15/propositions/pion0965.asp
    if not all_articles:
        for rejected in rejected_all_articles:
            articles_parsed = [art for art in rejected if art.get('type') == 'article']
            if len(articles_parsed):
                print('WARNING: retrieving parsed text from a previously rejected text')
                all_articles = rejected
                break

    if article is not None:
        pr_js(article)

    if indextext != -1 and curtext + 1 != len(srclst):
        print("WARNING: multiple texts announced but %d/%d found %s" % (curtext + 1, len(srclst), srclst), indextext)

    return all_articles



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re, os, sys
from difflib import ndiff
from functools import cmp_to_key

from tlfp.tools import _step_logic
from tlfp.tools.common import open_json, print_json, clean_text_for_diff, compute_similarity, format_display_date
from tlfp.tools.sort_articles import compare_articles


def getParentFolder(root, f):
    abs = os.path.abspath(os.path.join(root, f))
    return os.path.basename(os.path.abspath(os.path.join(abs, os.pardir)))


def unifyStatus(status):
    status = status.lstrip().rstrip('s. ')
    if status.endswith('constitution') or status.startswith('sup'):
        return "sup"
    if status.startswith("nouveau"):
        return "new"
    return "none"


def create_step(step_id, article=None, echec_type=None):
    s = {}
    s['id_step'] = step_id
    s['directory'] = step_id
    s['text'] = []
    if article:
        if article.get('statut'):
            s['status'] = unifyStatus(article['statut'])
        else:
            s['status'] = 'none'
        for key in sorted(article['alineas'].keys()):
            if article['alineas'][key] != '':
                s['text'].append(article['alineas'][key])
    else:
        s['status'] = echec_type.upper()
        s['length'] = 0
        s['n_diff'] = 0
    return s


def mark_missing_articles_as_deleted(articles, old_step_id, step_id, last_match_with_previous_step, current_match):
    # we look for *non-deleted* articles in the previous step to mark them as deleted in this step
    # articles to recover = articles with index between last_match_with_previous_step and current_match
    for article_id, article in sorted(articles.items()):
        for step in article['steps']:
            if step['id_step'] == old_step_id \
                    and last_match_with_previous_step < step['_original_index'] < current_match \
                    and step['status'] != 'sup':
                print('Matched an article deleted in this step', article_id)
                def_s = dict(step)
                def_s['id_step'] = step_id
                def_s['directory'] = step_id
                def_s['status'] = 'sup'
                def_s['length'] = 0
                def_s['diff'] = 'rem'
                def_s['n_diff'] = 0
                articles[article_id]['steps'].append(def_s)


def check_text_match_dosleg(url, senat_id):
    # test if the text is the one 
    if url.endswith('/%s.html' % senat_id):
        return True


    # test the short version: pjl2014-211 -> pjl14-211
    if len(senat_id) > len('pjlYY-XXX') and senat_id.startswith('pjl20'): # TODO: update in one century :)
        senat_id = senat_id.replace('pjl20', 'pjl') 
        if url.endswith('/%s.html' % senat_id):
            return True

    return False


def reordered_steps(procedure):
    # Handle reorder of repeated depots (typically a few PPL from Senat similar to a PJL added to its dossier)
    senat_id = procedure.get('senat_id')
    first = None
    steps = []
    latersteps = []
    for i, step in enumerate(procedure['steps']):
        if step.get('step', '') == 'depot':
            if not first and (step.get('institution', '') == 'assemblee' or check_text_match_dosleg(step.get('source_url', ''), senat_id)):
                first = step
            elif first and step.get('institution', '') == 'assemblee' and check_text_match_dosleg(step.get('source_url', ''), senat_id):
                continue
            elif step.get('institution', '') == 'senat' and "/ppl" in step.get('source_url', '') and step.get('stage', '') == '1ère lecture':
                # check next step is a depot too
                if len(procedure['steps']) > i + 1 and procedure['steps'][i + 1].get('step') == 'depot':
                    steps.append(step)
            else:
                continue
        else:
            latersteps.append(step)
    steps.append(first)
    depots = len(steps)
    steps += latersteps
    return depots, steps


def process(procedure):
    title = procedure.get("long_title", "Missing title").replace(procedure.get("short_title", "").lower(), procedure.get("short_title", ""))
    title = title[0].upper() + title[1:]
    out = {'law_title': title, 'articles': {}, 'sections': {}, 'short_title': procedure.get("short_title", "")}
    if 'loi_dite' in procedure:
        out['loi_dite'] = procedure['loi_dite']

    depots, steps = reordered_steps(procedure)

    # TODO : check whether it should be reintegrated or not
    # skip step presently happening
    # if not steps[-1].get('enddate'):
    #   steps.pop(-1)

    re_upper_first = re.compile(r'^(.)(.*)$')
    step_id = ''
    old_step_index = None
    for nstep, step in enumerate(steps):
        data = step.get('texte.json')
        if step['stage'] == 'promulgation':
            continue
        if not data and not step.get('echec'):
            if not ((procedure.get('is_old_procedure') or _step_logic.use_old_procedure(step)) and step.get('step') == 'commission'):
                print('       WARNING: prepare_articles: no data for', step.get('stage'), step.get('step'), step.get('institution'), file=sys.stderr)
            continue

        step_id = step['directory']
        print('      * preparing articles for step', step_id)

        # hack
        step['echec'] = step.get('echec')

        if step['echec']:
            if 'echec' not in out['articles']:
                out['articles']['echec'] = {'id': 'echec', 'titre': step['echec'], 'section': 'echec', 'steps': []}
            next_step = create_step(step_id, echec_type=step['echec'])
            out['articles']['echec']['steps'].append(next_step)
            if 'echec' not in out['sections']:
                out['sections']['echec'] = {}
            if step['echec'] == 'renvoi en commission' and not data:
                data = {'expose': 'Siégeant en hémicycle le %s, les %s ont adopté une motion de renvoi du texte en commission.' % (format_display_date(step.get('enddate', step['date'])), "sénateurs" if step['institution'] == "senat" else "députés")}
            out['sections']['echec'][step_id] = {'title': data['expose'] if data else '', 'type': step['echec'].upper()}
            continue
        for section in data['sections']:
            if not section['id'] in out['sections']:
                out['sections'][section['id']] = {}
            out['sections'][section['id']][step_id] = {'title': section['titre'], 'type': re_upper_first.sub(lambda x: x.group(1).upper() + x.group(2), section['type_section'])}
            if 'newid' in section:
                out['sections'][section['id']][step_id]['newnum'] = section['newid']
        last_match_with_previous_step = -1
        for article_index, article in enumerate(data['articles']):
            id = article['titre'].replace(' ', '_')
            if out['articles'].get(id):
                s = create_step(step_id, article=article)
                if 'newtitre' in article:
                    s['newnum'] = article['newtitre']
                txt = clean_text_for_diff(s['text'])

                oldtext = []
                if old_step_index is not None:
                    old_step_id = steps[old_step_index]['directory']
                    for st in out['articles'][id]['steps']:
                        if st['id_step'] == old_step_id:
                            if st['_original_index'] != last_match_with_previous_step + 1:
                                mark_missing_articles_as_deleted(out['articles'], old_step_id, step_id, last_match_with_previous_step, st['_original_index'])
                            last_match_with_previous_step = st['_original_index']
                            if st['status'] != 'sup':
                                oldtext = st['text']
                            break

                if txt and (not oldtext or nstep < depots):
                    s['status'] = 'new' if nstep >= depots else 'none'
                    s['diff'] = 'add'
                    s['n_diff'] = 1
                elif s['status'] == "sup":
                    s['diff'] = 'rem'
                    s['n_diff'] = 0
                else:
                    oldtxt = clean_text_for_diff(oldtext)
                    s['status'] = 'none'
                    if txt == oldtxt:
                        s['diff'] = 'none'
                        s['n_diff'] = 0
                    else:
                        compare = list(ndiff(s['text'], oldtext))
                        mods = {'+': 0, '-': 0}
                        for line in compare:
                            mod = line[0]
                            if mod not in mods:
                                mods[mod] = 0
                            mods[mod] += 1
                        if mods['+'] > mods['-']:
                            s['diff'] = 'add'
                        elif mods['+'] < mods['-']:
                            s['diff'] = 'rem'
                        elif mods['+'] * mods['-']:
                            s['diff'] = 'both'
                        else:
                            s['diff'] = 'none'
                        s['n_diff'] = 1 - compute_similarity(oldtxt, txt)
            else:
                out['articles'][id] = {}
                out['articles'][id]['id'] = id
                out['articles'][id]['titre'] = article['titre']
                if article.get('section'):
                    out['articles'][id]['section'] = article['section']
                else:
                    out['articles'][id]['section'] = 'A%s' % article['titre']
                out['articles'][id]['steps'] = []
                s = create_step(step_id, article)
                s['n_diff'] = 1
                s['diff'] = 'add'
                if nstep >= depots:
                    s['status'] = 'new'
                else:
                    s['status'] = 'depot'
            if s['status'] == 'sup':
                s['length'] = 0
                s['n_diff'] = 0
            else:
                s['length'] = len("\n".join(s['text']))
            s['_original_index'] = article_index
            out['articles'][id]['steps'].append(s)
        old_step_index = nstep

        # except Error as e:
        #     sys.stderr.write("ERROR parsing step %s:\n%s: %s\n" % (str(step)[:50], type(e), e))
        #    exit(1)

    for a in sorted(out['articles']):
        new_steps = []
        for s in out['articles'][a]['steps']:
            del s['text']
            s.pop('_original_index', None)
            if len(new_steps) > 0 and new_steps[-1]['id_step'] == s['id_step']:
                print('same id_step', s['id_step'], file=sys.stderr)
                continue
            new_steps.append(s)
        out['articles'][a]['steps'] = new_steps

    # Set articles' order values after having reinserted missing ones
    orders = {k: n for n, k in enumerate(
        sorted(
            [a['titre'] for a in out['articles'].values() if a['id'] != 'echec'],
            key=cmp_to_key(compare_articles)
        ))
    }
    for a in out['articles'].values():
        if a['id'] == 'echec':
            a['order'] = -1
        else:
            a['order'] = orders[a['titre']]

    return out


if __name__ == '__main__':
    print_json(process(open_json(sys.argv[1])))



import os
import sys
import re
from time import time
from functools import cmp_to_key

from lawfactory_utils.urls import download

from .common import Context, open_json, get_text_id, \
    identify_room, print_json, amdapi_link, \
    national_assembly_text_legislature
from .sort_articles import compare_articles
from ._step_logic import get_previous_step


def process(OUTPUT_DIR, procedure):
    context = Context(OUTPUT_DIR, load_parls=True)

    #['Indéfini', 'Adopté', 'Irrecevable', 'Rejeté', 'Retiré', 'Tombe', 'Non soutenu', 'Retiré avant séance', 'Rectifié', 'Favorable' ,'Satisfait']
    def simplify_sort(sort):
        sort = sort.lower()
        if sort in "adopté favorable":
            return "adopté"
        if sort in "rejeté ":
            return "rejeté"
        if sort in "indéfini":
            return "en attente"
        return "non-voté"

    re_clean_first = re.compile(r'^(.*?)(,| et) .*$')
    def first_author(signataires):
        if signataires is None or "gouvernement" in signataires.lower():
            return ""
        return re_clean_first.sub(r'\1, …', signataires)

    def find_groupe(amd, typeparl, urlapi):
        if amd['signataires'] and "gouvernement" in amd['signataires'].lower():
            return "Gouvernement"

        # Fix groupes not historicized in NosSénateurs
        if typeparl == "senateur" and amd["parlementaires"]:
            return context.get_senateur_groupe(amd["parlementaires"][0]["parlementaire"], amd["date"], urlapi)

        return amd['auteur_groupe_acronyme']

    def add_link(links, pA, pB, weight=1):
        p1 = min(pA, pB)
        p2 = max(pA, pB)
        linkid = "%s-%s" % (p1, p2)
        if linkid not in links:
            links[linkid] = {
              "1": p1,
              "2": p2,
              "w": 0
            }
        links[linkid]["w"] += weight

    article_number_regexp = re.compile(r'article (1er.*|(\d+).*)$', re.I)
    def sort_amendements(texte, amendements):
        articles = {}
        for article in texte:
            if article['type'] == 'article':
                titre = article.get('titre')
                if titre:
                    articles[titre.lower()] = article.get('order') * 10

        def solveorder(art):
            nonlocal articles
            art = art.lower()
            order = 10000
            if art == 'titre' or art.startswith('intitul'):
                return 0
            elif art.startswith('motion'):
                return 1
            elif art.startswith('projet') \
                or art.startswith('proposition') \
                or art.startswith('texte'):
                return 5
            else:
                m = article_number_regexp.search(art)
                if m:
                    if articles.get(m.group(1)):
                        order = articles.get(m.group(1))
                    elif articles.get(m.group(2)):
                        order = articles.get(m.group(2))
                    if 'avant' in art:
                        order -= 1
                    elif 'après' in art or 'apres' in art:
                        order += 1
            return order


        for amendement in amendements:
            amdt = amendement['amendement']
            amdt['ordre_article'] = solveorder(amdt['sujet'])

        return amendements


    CACHE_BUSTING = 'cache=%d' % time()
    if 'url_jo' in procedure:
        CACHE_BUSTING = 'cache=lfdll-prod' # fixed cache busting for promulgated laws
    steps = {}
    last_text_id, last_text_typeparl = None, None
    steps = procedure['steps']
    for i, step in enumerate(steps):
        print('    * step -', step.get('stage'), step.get('step'), step.get('source_url'))
        if step.get('step') not in ('commission', 'hemicycle'):
            continue
        if step.get('step') == 'commission' and step.get('stage') == 'CMP':
            continue

        if i == 0:
            continue

        last_step_index = get_previous_step(steps, i, is_old_procedure=procedure.get('use_old_procedure'))
        last_step = steps[last_step_index]
        last_step_with_good_text_number = steps[get_previous_step(steps, i,
            is_old_procedure=procedure.get('use_old_procedure'), get_depot_step=True)
        ]
        texte_url = last_step_with_good_text_number.get('source_url')

        if last_step.get('in_discussion'):
            print('WARNING: ignoring future steps further than current discussion', file=sys.stderr)
            break

        if step.get('stage') != 'CMP' and last_step_with_good_text_number.get('institution') != step.get('institution'):
            print('ERROR - last step is from another institution', file=sys.stderr)
            continue

        # for a CMP hemicycle we have to get the right text inside the CMP commission
        if step.get('stage') == 'CMP' and step.get('step') == 'hemicycle':
            urls = [last_step.get('source_url')]
            if 'cmp_commission_other_url' in last_step:
                urls.append(last_step.get('cmp_commission_other_url'))
            an_url = [url for url in urls if 'nationale.fr' in url]
            senat_url = [url for url in urls if 'senat.fr' in url]
            if step.get('institution') == 'assemblee' and an_url:
                texte_url = an_url[0]
            elif step.get('institution') == 'senat' and senat_url:
                texte_url = senat_url[0]
            else:
                print('WARNING - missing the CMP commission text for', step.get('source_url'), file=sys.stderr)
                continue

        if texte_url is None:
            print('ERROR - no texte url', step.get('source_url'), file=sys.stderr)
            continue

        legislature = None
        if 'assemblee-nationale.fr' in texte_url:
            legislature = national_assembly_text_legislature(texte_url)

        texte = open_json(os.path.join(context.sourcedir, 'procedure', last_step['directory']), 'texte/texte.json')

        typeparl, urlapi = identify_room(texte_url, legislature)

        amdt_url = None
        if "nationale.fr" in texte_url:
            if 'assemblee_legislature' not in procedure:
                print('         + no AN legislature - pass text')
                continue
            amdt_url = 'https://%s.fr/%s/amendements/%s/json?%s' % (urlapi, legislature, get_text_id(texte_url), CACHE_BUSTING)
        elif "senat.fr" in texte_url:
            amdt_url = 'https://%s.fr/amendements/%s/json?%s' % (urlapi, get_text_id(texte_url), CACHE_BUSTING)

        if amdt_url is None:
            continue

        print('      * downloading amendments:', amdt_url, 'for', texte_url)

        try:
            amendements_src = download(amdt_url).json().get('amendements', [])
        except:
            raise Exception("ERROR: amendements JSON at %s is badly formatted, it should probably be hardcached on ND/NS" % amdt_url)

        # TA texts can be zero-paded or not (TA0XXX or TAXXX), we try both
        if 'amendements/TA' in amdt_url:
            textid = get_text_id(texte_url)
            if 'TA0' in textid:
                alternative_url = amdt_url.replace(textid, 'TA' + textid.replace('TA', '').lstrip('0'))
            else:
                alternative_url = amdt_url.replace(textid, 'TA' + textid.replace('TA', '').zfill(4))
            print(' WARNING: TA - trying alternative url too', alternative_url)
            try:
                amendements_src += download(alternative_url).json().get('amendements', [])
            except:
                raise Exception("ERROR: amendements JSON at %s is badly formatted, it should probably be hardcached on ND/NS" % alternative_url)

        print('        parsing amendments:', len(amendements_src))

        # ignore amendments if they are not for the correct step
        amendements_src_filtered = []
        for amd in amendements_src:
            a = amd['amendement']
            if step.get('institution') == 'assemblee':
                # commission amendments can have two forms
                #    - /amendements/LOI/NUM.asp (13th legislature)
                #    - /amendements/LOI/COMMISSION_NAME/NUM.asp (14+ legislature)
                # hemicycle amendments are:
                #    - /amendements/LOI/NUM.asp (13th legislature)
                #    - /amendements/LOI/AN/NUM.asp (14+ legislature)
                amdt_step = 'hemicycle'
                if '/cr-' in a['source']:
                    amdt_step = 'commission'
                else:
                    url_parts = a['source'].split('amendements/')[1].split('/')
                    if len(url_parts) == 3 and url_parts[1] != 'AN':
                        amdt_step = 'commission'
            elif step.get('institution') == 'senat':
                amdt_step = 'commission' if '/commissions/' in a['source'] else 'hemicycle'
            else:
                # CMP - there's not way for now to distinguish the step
                amdt_step = step['step']
            if step['step'] != amdt_step:
                continue
            amendements_src_filtered.append(amd)

        if len(amendements_src_filtered) != len(amendements_src):
            print('WARNING: amendments ignored (not the right step) %s' %
                    (len(amendements_src) - len(amendements_src_filtered)), file=sys.stderr)
        amendements_src = amendements_src_filtered

        step['nb_amendements'] = len(amendements_src)

        if len(amendements_src) > 0:
            amendements_src = sort_amendements(texte['articles'], amendements_src)

            sujets = {}
            groupes = {}

            fix_order = False
            orders = []
            parls = {}
            links = {}
            idents = {}
            for amd in amendements_src:
                a = amd['amendement']
                if "sort" not in a:
                    print('WARNING: amendment has no sort %s\n' % a['url_nos%ss' % typeparl], file=sys.stderr)
                    continue
                if a["sort"] == "Rectifié":
                    continue
                if "sujet" not in a or not a["sujet"]:
                    if a["sort"] not in ["Irrecevable", "Retiré avant séance"]:
                        print('WARNING: amendment has no subject %s\n' % a['url_nos%ss' % typeparl], file=sys.stderr)
                    continue
                key = a['sujet']
                if not key:
                    print('WARNING: amendment has no subject %s\n' % a['url_nos%ss' % typeparl], file=sys.stderr)
                    continue
                if key not in sujets:
                    orders.append(key)
                    sujets[key] = {
                      'titre': key,
                      'order': a['ordre_article'],
                      'amendements': []
                    }
                if a['ordre_article'] > 9000:
                    fix_order = True

                gpe = find_groupe(a, typeparl, urlapi)
                if not gpe:
                    if a["sort"] != "Irrecevable":
                        sys.stderr.write('WARNING: no groupe found for %s\n' % a['url_nos%ss' % typeparl])
                    gpe = "Inconnu"
                context.add_groupe(groupes, gpe, urlapi)

                sujets[key]['amendements'].append({
                  'numero': a['numero'],
                  'date': a['date'],
                  'sort': simplify_sort(a['sort']),
                  'groupe': gpe,
                  'id_api': a['id'],
                  'aut': first_author(a['signataires'])
                })

                cosign = []
                hmd5 = a["cle_unicite"]
                if hmd5 not in idents:
                    idents[hmd5] = []
                for parll in a["parlementaires"]:
                    parl = parll["parlementaire"]
                    if parl not in parls:
                        p = context.get_parlementaire(urlapi, parl)
                        parls[parl] = {
                          "i": p["id"],
                          "s": parl,
                          "a": 0,
                          "n": p["nom"],
                          "g": p["groupe_sigle"],
                          "p": p["place_en_hemicycle"]
                        }
                    pid = parls[parl]["i"]
                    parls[parl]["a"] += 1
                    for cid in cosign:
                        add_link(links, pid, cid)
                        #add_link(links, pid, cid, 2)
                    cosign.append(pid)
                    for cid in idents[hmd5]:
                        add_link(links, pid, cid)
                    idents[hmd5].append(pid)

            if fix_order:
                orders.sort(key=cmp_to_key(compare_articles))
                for i, k in enumerate(orders):
                    sujets[k]["order"] = i

            amdtsfile = os.path.join(context.sourcedir, 'viz', 'amendements_%s.json' % step['directory'])
            data = {'id_step': step['directory'],
                    'api_root_url': amdapi_link(urlapi),
                    'groupes': groupes,
                    'sujets': sujets}
            print_json(data, amdtsfile)

            linksfile = os.path.join(context.sourcedir, 'viz', 'amendements_links_%s.json' % step['directory'])
            data = {'id_step': step['directory'],
                    'links': list(links.values()),
                    'parlementaires': dict((p["i"], dict((k, p[k]) for k in "psang")) for p in list(parls.values()))}
            # print_json(data, linksfile)

        ###########  INTERVENTIONS #############
        # TODO: move this to a dedicated file

        print('      * downloading interventions')
        typeparl, urlapi = identify_room(texte_url, legislature)
        inter_dir = os.path.join(context.sourcedir, 'procedure', step['directory'], 'interventions')
        commission_or_hemicycle = '?commission=1' if step.get('step') == 'commission' else '?hemicycle=1'
        # TODO: TA texts can be zero-paded or not (TA0XXX or TAXXX), we should try both
        seance_name = None
        intervention_files = []

        texts = (get_text_id(texte_url),)
        if last_text_typeparl == typeparl:
            texts = (get_text_id(texte_url), last_text_id)

        for loiid in texts:
            if typeparl == 'depute':
                url_seances = 'https://%s.fr/%s/seances/%s/json%s' % (urlapi, legislature, loiid, commission_or_hemicycle)
            else:
                url_seances = 'https://%s.fr/seances/%s/json%s' % (urlapi, loiid, commission_or_hemicycle)

            print('        * downloading seances - ', url_seances)
            for id_seance_obj in sorted(download(url_seances).json().get('seances', []), key=lambda x: x["seance"]):
                if typeparl == 'depute':
                    url_seance = 'https://%s.fr/%s/seance/%s/%s/json' % (urlapi, legislature, id_seance_obj['seance'], loiid)
                else:
                    url_seance = 'https://%s.fr/seance/%s/%s/json' % (urlapi, id_seance_obj['seance'], loiid)

                print('           downloading seance - ', url_seance)
                resp = download(url_seance).json()
                if resp.get('seance'):
                    inter = resp.get('seance')[0]['intervention']
                    seance_name = inter['date'] + 'T' + inter['heure'] + '_' + inter['seance_id']
                    print('            dumping seance -', seance_name)
                    intervention_files.append(seance_name)
                    if not os.path.exists(inter_dir):
                        os.makedirs(inter_dir)
                    print_json(resp, os.path.join(inter_dir, seance_name + '.json'))
            if seance_name:
                step['has_interventions'] = True
                step['intervention_files'] = intervention_files
                break

        last_text_id = get_text_id(texte_url)
        last_text_typeparl = typeparl

    return procedure


if __name__ == '__main__':
    process(sys.argv[1], open_json(os.path.join(sys.argv[1], 'viz/procedure.json')))



"""
Usage: python generate_dossiers_csv.py <api_directory>

Output in <api_directory>:
- dossiers_promulgues.csv with all the doslegs ready
- home.json for the homepage informations
"""
import glob, os, sys, csv, re, copy, datetime

from tlfp.tools.common import upper_first, open_json, print_json

API_DIRECTORY = sys.argv[1]

re_dos_ok = re.compile(r"%s/[^.]+/" % API_DIRECTORY.strip('/'))
dossiers = [(open_json(path), path) for path \
                in glob.glob(os.path.join(API_DIRECTORY, '*/viz/procedure.json')) if re_dos_ok.search(path)]
dossiers = [(dos, path) for dos, path in dossiers if "_tmp" not in path]


csvfile = csv.writer(open(os.path.join(API_DIRECTORY, 'dossiers.csv'), 'w'), delimiter=';')
csvfile.writerow(('id;Titre;Type de dossier;Date initiale;URL du dossier;État du dossier;Décision du CC;'
    'Date de la décision;Date de promulgation;Numéro de la loi;Thèmes;total_amendements;total_mots;'
    'short_title;loi_dite;assemblee_id').split(';'))


def format_date_for_human(date):
    if not date:
        return ''
    return '/'.join(reversed(date.split('-')))

def in_room(step):
    if step.get('institution') == "senat":
        return "au Sénat"
    return "à l'Assemblée"

def format_statuses(dos):
    status_amendements = ''
    if dos['stats']['total_amendements'] == 0:
        status_amendements = 'Aucun amendement'
    elif dos['stats']['total_amendements'] == 1:
        status_amendements = 'Un amendement'
    else:
        status_amendements = '%d amendements' % dos['stats']['total_amendements']

    status_live = ''
    if not dos.get('url_jo'):
        in_discussion_step = [step for step in dos['steps'] if step.get('in_discussion')]
        if in_discussion_step:
            in_discussion_step = in_discussion_step[0]
            if in_discussion_step.get('step') in ('commission', 'hemicycle') and in_discussion_step.get('date'):
                today_date = datetime.date.today().strftime(r'%Y-%m-%d')
                date = in_discussion_step.get('date')
                room = in_room(in_discussion_step)
                if date > today_date:
                    status_live = "à l'ordre du jour %s le %s" % (room, format_date_for_human(date))
                elif date == today_date:
                    status_live = "à l'ordre du jour %s aujourd'hui" % (room, )
                else:
                    status_live = "dernière discussion %s le %s" % (room, format_date_for_human(date))

        if not status_live:
            last_step = [step for step in dos['steps'] if step.get('date') and step.get('debats_order') is not None]
            if last_step and last_step[-1].get('date'):
                last_step = last_step[-1]
                date = format_date_for_human(last_step.get('enddate') or last_step.get('date'))
                room = in_room(last_step)
                if last_step.get('step') == 'depot':
                    status_live = "déposé %s le %s" % (room, date)
                elif last_step.get('step') in ('commission', 'hemicycle'):
                    status_live = "dernière discussion %s le %s" % (room, date)

    year = dos.get('end').split('-')[0] if dos.get('end') else ''

    return {
        'live': status_live,
        'recent': 'promulgué le %s' % format_date_for_human(dos.get('end')),
        'focus': status_amendements + ' (%s)' % year,
    }


def select_status(data, status_key):
    """
    We store the multiple possible statuses in the status attribute.
    This fonction select the right one for each column of the homepage
    """
    final_data = []
    for dos in copy.deepcopy(data):
        dos['status'] = dos['status'][status_key]
        final_data.append(dos)
    return final_data


home_json_data = []

total_doslegs = total_promulgues = 0
for dos, path in dossiers:
    if not dos.get('beginning'):
        print('INVALID BEGGINING DATE:', dos['id'])
        continue

    decision_cc = None
    decision_cc_date = None
    for step in dos['steps']:
        if step.get('stage') == 'constitutionnalité':
            decision_cc = step.get('decision')
            decision_cc_date = step.get('date')
            break

    csvfile.writerow([
        dos['id'], # id
        dos.get('long_title'), # Titre
        'proposition de loi' if dos.get('proposal_type') == 'PPL' else 'projet de loi', # Type de dossier
        dos.get('beginning'), # Date initiale
        dos.get('url_dossier_senat', dos.get('url_dossier_assemblee')), # URL du dossier
        'promulgué' if dos.get('end') else '', # État du dossier
        decision_cc, # Décision du CC
        decision_cc_date, # Date de la décision
        dos.get('end'), # Date de promulgation
        dos.get('law_name'), # Numéro de la loi
        ','.join(dos.get('themes', [])), # Thèmes
        dos['stats']['total_amendements'], # total_amendements
        dos['stats']['total_mots'], # total_mots
        dos.get('short_title'), # short_title
        dos.get('loi_dite'), # Nom commun de la loi
        dos.get('assemblee_id'), # ex: 14-peche_et_chasse
    ])

    # format data for home.json
    if dos['stats']['total_amendements']:
        title = dos.get('short_title')
        if dos.get('loi_dite'):
            title = "%s (%s)" % (upper_first(dos.get('loi_dite')), title)

        maxdate = dos.get('end')
        if not maxdate:
            for step in dos['steps']:
                if step.get('date') and step.get('debats_order') is not None:
                    maxdate = step.get('enddate') or step.get('date')

        home_json_data.append({
            'total_amendements': dos['stats']['total_amendements'],
            'end': dos.get('end'),
            'maxdate': maxdate,
            'status': format_statuses(dos),
            'loi': dos['id'],
            'titre': title
        })

    total_doslegs += 1
    if dos.get('url_jo'):
        total_promulgues += 1

total_encours = total_doslegs - total_promulgues

erreurs = len(glob.glob(os.path.join(API_DIRECTORY, 'logs/*')))
erreurs_encours = len(glob.glob(os.path.join(API_DIRECTORY, 'logs-encours/*')))

max_promulgues = total_promulgues + erreurs
max_encours = total_encours + erreurs_encours
maximum = max_promulgues + max_encours

print(total_doslegs, 'doslegs in csv')
print('%.1f%s (%d/%d)' % (100*total_promulgues/max_promulgues, '%', total_promulgues, max_promulgues), 'de promulgués qui passent')
print('%.1f%s (%d/%d)' % (100*total_encours/max_encours, '%', total_encours, max_encours), 'de textes en cours qui passent')


#### Make home.json

home_json_final = {
    "total": total_promulgues,
    "encours": total_encours,
    "maximum": max_promulgues,
}

TEXTS_PER_COLUMN = 6
MIN_AMENDMENTS = 50

most_amended = [dos for dos in sorted(home_json_data, key=lambda x: -x['total_amendements']) if dos['end']]
home_json_final["focus"] = {
    "titre": "Les textes les plus amendés",
    "lien": "Explorer les textes les plus amendés",
    "url": "lois.html?action=quanti",
    "textes": select_status(most_amended[:TEXTS_PER_COLUMN], "focus"),
}

recent = sorted(home_json_data, key=lambda x: x['end'] or "0")
recent = [dos for dos in reversed(recent) if dos['end'] and dos['total_amendements'] >= 50]
recent = list(recent)
home_json_final["recent"] = {
    "titre": "Les derniers textes promulgués",
    "lien": "Explorer les textes récents",
    "url": "lois.html",
    "textes": select_status(recent[:TEXTS_PER_COLUMN], "recent"),
}

live = sorted(home_json_data, key=lambda x: x['maxdate'])
live = [dos for dos in reversed(live) if not dos['end'] and dos['total_amendements'] >= 50]
live = list(live)
home_json_final["live"] = {
    "titre": "Les textes en cours",
    "lien": "Explorer les textes en cours",
    "url": "lois.html?encours",
    "textes": select_status(live[:TEXTS_PER_COLUMN], "live"),
}

print_json(home_json_final, os.path.join(API_DIRECTORY, 'home.json'))
print('home.json OK')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
from functools import cmp_to_key

# Handle all bis-ter words until 27 included
# (adding 28 duodetrecies creates more complexity)
# cf http://fr.wikipedia.org/wiki/Adverbe_multiplicatif
# 27 is actually the biggest used case in all texts analyzed so far back to 2009
# cf http://www.assemblee-nationale.fr/13/ta-commission/r3604-a0.asp
# TODO: support multiple spellings in tests + more values than continuous
#       ex: quatervecies/quatervicies, novies/nonies
bis_27 = ['bis', 'ter', 'quater', 'quinquies', 'sexies', 'septies', 'octies', 'novies',
'decies', 'undecies', 'duodecies', 'terdecies', 'quaterdecies', 'quindecies', 'sexdecies', 'septdecies', 'octodecies', 'novodecies',
'vicies', 'unvicies', 'duovicies', 'tervicies', 'quatervicies', 'quinvicies', 'sexvicies', 'septvicies', 'duodetrecies', 'undetricies', 'tricies']

# support 1 to 99, from https://framagit.org/parlement-ouvert/metslesliens/blob/master/docs/l%C3%A9gistique.md
bister = '(' + \
  '(?:un|duo|ter|quater|quin|sex?|sept|octo|novo|unde?|duode)?' + \
  '(?:dec|v[ie]c|tr[ie]c|quadrag|quinquag|sexag|septuag|octog|nonag)' + \
  'ies|semel|bis|ter|quater|' + \
  '(?:quinqu|sex|sept|oct|no[nv])ies' + \
  ')'
re_bister = re.compile(bister)

re_article = re.compile(r'(\d+)e?r?(( ([A-Z]+|%s))*)' % bister)

re_clean_befaft = re.compile(r"^(a(vant|près)\s*l'|article\s*)+", re.I)

re_add_spaces = re.compile(r'([A-Z])\s*')
add_spaces = lambda x: re_add_spaces.sub(r'\1 ', x)
def split_article(a):
    if not re_article.match(re_clean_befaft.sub('', a)):
        return [0, a]
    m = re_article.search(a)
    res = [int(m.group(1))]
    if m.group(2):
        res += add_spaces(m.group(2)).strip().split(' ')
    return res

hash_bis = {'u': 1, 'b': 2, 't': 3, 'o': 8, 'n': 9,
  'qua': 4, 'qui': 5, 'sex': 6, 'sep': 7, 'du': 2, 'de': 0, 'tri': 0, 'v': 0}
def quantify_bis(b):
    u = 0
    if b.startswith("und") and b != 'undecies':
        u = -1
    elif b.startswith("duode") and b != 'duodecies':
        u = -2
    else:
        for i in [3, 2, 1]:
            if b[:i] in hash_bis:
                u = hash_bis[b[:i]]
                break
    if 'decies' in b:
        u += 10
    elif 'vicies' in b or 'vecies' in b:
        u += 20
    elif 'tricies' in b or 'trecies' in b:
        u += 30
    elif 'quadragies' in b:
        u += 40
    elif 'quinquagies' in b:
        u += 50
    elif 'sexagies' in b:
        u += 60
    elif 'septuagies' in b:
        u += 70
    elif 'octogies' in b:
        u += 80
    elif 'nonagies' in b:
        u += 90
    return u

def type_detail(d):
    if not d:               # nothing
        return 0
    if len(d) == 1:         # A-Z
        return -1
    if re_bister.match(d):  # bis-ter
        return 1
    return -2               # junk text

def compare_details(a, b):
    if a == b:
        return 0
    # recursive if same first detail
    if a[0] == b[0]:
        return compare_details(a[1:], b[1:])
    # classify type details
    qa = type_detail(a[0])
    qb = type_detail(b[0])
    # order by A-Z < _ < bis-ter
    if qa != qb:
        return qa - qb
    # if both A-Z or junk text
    if qa < 0:
        if a[0] < b[0]: # use alpha order
            return -1
        return 1
    # necessarily both bis-ter now
    return quantify_bis(a[0]) - quantify_bis(b[0])

def compare_articles(a, b):
    if a == b:
        return 0
    na = split_article(a)
    nb = split_article(b)
    # compare numbers
    if na[0] != nb[0]:
        return na[0] - nb[0]
    # equalize array lengths
    if len(na) > len(nb):
        nb += [None] * (len(na) - len(nb))
    elif len(na) < len(nb):
        na += [None] * (len(nb) - len(na))
    # compare details
    res = compare_details(na[1:], nb[1:])
    if res != 0:
        return res
    ia = na[0]
    if 'avant' in a.lower():
        ia -= 1
    elif 'après' in a.lower():
        ia += 1
    ib = nb[0]
    if 'avant' in b.lower():
        ib -= 1
    elif 'après' in b.lower():
        ib += 1
    return ia - ib

def article_is_lower(a, b):
    return compare_articles(a, b) < 0

if __name__ == "__main__":
    # Test split articles
    print("[TEST] Splitting article 1er A bis AA'")
    assert(split_article('1er A bis AA') == [1, 'A', 'bis', 'A', 'A'])
    print(" -> Success!")

    print("[TEST] Splitting article 1er A duodetrecies AA'")
    assert(split_article('1er A duodetrecies AA') == [1, 'A', 'duodetrecies', 'A', 'A'])
    print(" -> Success!")

    # Test convert bis to numbers for 2 to 27
    print("[TEST] Converting bis expressions to numbers for 2 to 27:")
    for i, v in enumerate(bis_27):
        try:
            assert(quantify_bis(v) == i + 2)
        except Exception as e:
            print("Wrong quantifying of '%s': %s" % (v, quantify_bis(v)))
            raise e
    print(" -> Success!")

    # Test sorting an array of articles
    print("[TEST] Sorting randomized array of articles:")
    sorted_arts = [
      "liminaire",
      "wrong name 1",
      "wrong name 2",
      "1er A",
      "1er A bis AA",
      "1er A bis A",
      "1er A bis",
      "1er B",
      "Après l'article 1er B",
      "Avant l'article 1er C",
      "1er C",
      "Après l'article 1er D",
      "avant l'article 1er",
      "1er",
      "Avant l'article 10 quater",
      "13",
      "14 AAA",
      "14 AA",
      "14 AB",
      "14 A",
      "14 A bis A",
      "14 A ter",
      "14 B",
      "14 CA",
      "14 C",
      "14",
      "14 bis A",
      "14 bis",
      "14 ter",
      "14 quater",
      "14 duodecies CA",
      "14 duodecies CC",
      "14 duodecies C",
      "14 duodecies G",
      "14 duodecies ZY",
      "14 duodecies ZZ",
      "14 duodecies Z",
      "14 duodecies",
    ]
    from pprint import pprint
    import random

    random_arts = list(sorted_arts)
    random.shuffle(random_arts)
    new_sorted_arts = list(sorted(random_arts, key=cmp_to_key(compare_articles)))

    if new_sorted_arts != sorted_arts:
        print("- Randomized array:")
        pprint(new_sorted_arts)
        print("- Sorted array:")
        pprint(sorted_arts)
    assert new_sorted_arts == sorted_arts
    print(" -> Success!")



"""
Make one Git repo per bill and optionally upload it to gitlab

TODO: ajouter acteurs manquants sur le gitlab
TODO: date
TODO: source_url
TODO: url gitlab as configurable option
"""

"""
find .procedure -name 'A*'  -type d | while read dir ; do echo mv '"'$dir'"' '"'$(echo $dir | sed 's|/A|/Article_|')'"' ; done |sh
find .procedure -name 'SS*' -type d | while read dir ; do echo mv '"'$dir'"' '"'$(echo $dir | sed 's|/SS|/SousSection_|')'"' ; done |sh
find .procedure -name 'S*'  -type d | while read dir ; do echo mv '"'$dir'"' '"'$(echo $dir | sed 's|/S|/Section_|')'"' ; done |sh
find .procedure -name 'C*'  -type d | while read dir ; do echo mv '"'$dir'"' '"'$(echo $dir | sed 's|/C|/Chapitre_|')'"' ; done |sh
find .procedure -name 'T*'  -type d | while read dir ; do echo mv '"'$dir'"' '"'$(echo $dir | sed 's|/T|/Titre_|')'"' ; done |sh
find .procedure -name 'L*'  -type d | while read dir ; do echo mv '"'$dir'"' '"'$(echo $dir | sed 's|/L|/Livre_|')'"' ; done |sh
find .procedure -name 'V*'  -type d | while read dir ; do echo mv '"'$dir'"' '"'$(echo $dir | sed 's|/V|/Volume_|')'"' ; done |sh
"""

import glob
import os
import sys
import shutil
import shlex
from pathlib import Path

import gitlab

from tlfp.tools.common import open_json


def read_text(path):
    # TODO: format tables
    try:
        articles = open_json(os.path.dirname(path), os.path.basename(path))["articles"]
    except FileNotFoundError:
        return ""

    texte = ""
    for art in articles:
        texte += "# Article " + art["titre"] + "\n\n"
        for key in sorted(art["alineas"].keys()):
            if art["alineas"][key] != "":
                texte += art["alineas"][key] + "\n"
        texte += "\n"
    return texte


def call(cmd):
    code = os.system(cmd)
    if code != 0:
        raise Exception('"{}"" returned {}'.format(cmd, code))


def call_bash(cmd):
    call("bash -c " + shlex.quote(cmd))


GIT_REPOS_DIRECTORY = sys.argv[1]


GITLAB_TOKEN = sys.argv[2] if len(sys.argv) == 3 else None
if GITLAB_TOKEN:
    gl = gitlab.Gitlab('https://git.regardscitoyens.org/', private_token=GITLAB_TOKEN)
    group = gl.groups.list(search='parlement')[0]

    # delete existing bills
    projects = group.projects.list()
    for project in projects:
        print('delete', project.id)
        gl.projects.delete(project.id)

for procedure_file in sorted(glob.glob("data/**/procedure.json", recursive=True)):
    procedure = open_json(procedure_file)

    if len(procedure["steps"]) < 5:
        continue
    if procedure["stats"]["total_amendements"] < 5:
        continue

    project_dir = os.path.dirname(os.path.dirname(procedure_file))

    git_dir = Path(GIT_REPOS_DIRECTORY) / procedure["id"]

    shutil.rmtree(str(git_dir), ignore_errors=True)
    os.makedirs(str(git_dir))

    remote_url = "git@git.regardscitoyens.org:/parlement/{bill}.git".format(
        bill=procedure["id"]
    )
    call_bash(
        "(cd %s;" % git_dir.absolute()
        + "git init;"
        + "git remote add origin %s)" % shlex.quote(remote_url)
    )

    prev_text = None
    for step in procedure["steps"]:
        # MSG=$(echo $line | awk -F ';' '{if ($8 == 1) print "Dépot du texte"; if ($8 != 1 && $11 != "depot" && $7 != "XX") print "Travaux en "$11", "$9;}');
        if step.get("step") == "depot":
            msg = "Dépot du texte"
        elif step.get("step") == "commission":
            msg = "Travaux en commission"
        elif step.get("step") == "hemicycle":
            msg = "Travaux en hemicycle"
        elif step.get("stage") == "promulgation":
            msg = "Promulgation"
        elif step.get("stage") == "constitutionnalité":
            msg = "Constitutionnalité"
        else:
            raise Exception("Unknown step: %s" % (str(step)))

        # AUTEUR=$(echo $line | awk -F ';' '{if ($8 == 1 && $6 ~ /pjl/) print "gouv"; else print $10;}');
        instit = step.get("institution")
        author_depot = step.get("auteur_depot")
        if instit == "gouvernement" or author_depot == "Gouvernement":
            author = "Gouvernement"
            author_email = "contact@pm.gouv.fr"
            gituser = "gouv"
        elif instit == "assemblee":
            author = author_depot or "Assemblée nationale"
            author_email = "contact@assemblee-nationale.fr"
            gituser = "assemblee"
        elif instit == "senat":
            author = author_depot or "Sénat"
            author_email = "contact@senat.fr"
            gituser = "senat"
        elif instit == "CMP":
            author = "Commission mixte paritaire"
            author_email = "contact@parlement.fr"
            gituser = "cmp"
        elif instit == "conseil constitutionnel":
            author = "Conseil Constitutionnel"
            author_email = " contact@conseil-constitutionnel.fr "
            gituser = "conseil_constitutionnel"
        elif instit == "congrès":
            author = "Congrès"
            author_email = "contact@assemblee-nationale.fr"  # ?
            gituser = "congrès"
        else:
            raise Exception("Unknown author: %s" % instit)

        # ID=$(echo $line | awk -F ';' '{print $7}')

        # DATE=$(echo $line | awk -F ';' '{print $14}')
        #  DATE=$(date --date="$DATE" -R);
        date = step.get("date")

        if "directory" not in step or step.get("debats_order") is None:
            continue

        texte_path = os.path.join(
            project_dir, "procedure", step.get("directory"), "texte/texte.json"
        )
        texte = read_text(texte_path)
        if texte:
            text_dest = git_dir / "texte"
            with open(str(text_dest), "w") as f:
                f.write(texte)

            #  echo $line | awk -F ';' '{print $2}' > texte/titre.txt

            r"""
            find texte -name *json -exec rm '{}' ';'
            find texte -name '*alineas' | sed 's/\(.*\)/mv "\1" "\1"/' | sed 's|[^/]*$|article.txt"|' | sh
            find texte -name 'article.txt' | while read dir; do echo mv '"'$dir'"' '"'$(echo $dir | sed 's|/article||')'"'; done | sh
            find texte -name '*titre' | sed 's/\(.*\)/mv "\1" "\1"/' | sed 's|[^/]*$|titre.txt"|' | sh
            find texte -size 0 -exec rm '{}' ';'
            """

            if prev_text == texte:
                msg += ' (aucun changement)'

            GIT_AUTHOR_NAME = GIT_COMMITER_NAME = author
            GIT_AUTHOR_EMAIL = GIT_COMMITER_EMAIL = author_email

            call_bash(
                "(cd %s" % git_dir.absolute()
                + "; git add *;"
                + "git status;"
                + "git config --local user.name %s;" % shlex.quote(GIT_AUTHOR_NAME)
                + "git config --local user.email %s;" % shlex.quote(GIT_AUTHOR_EMAIL)
                + " git commit --date=format:short:{} --author={} -m {} --allow-empty )".format(
                    shlex.quote(date),
                    shlex.quote(author + " <" + author_email + ">"),
                    shlex.quote(msg),
                )
            )
            prev_text = texte

    if GITLAB_TOKEN:
        try:
            gl.projects.create({
                'description': procedure['long_title'],
                'name': procedure['id'], # TODO: remove old AN projects
                'namespace_id': group.id,
                'visibility': 'public',
            })
        except gitlab.exceptions.GitlabCreateError:
            pass

        call_bash(
            "(cd %s;" % git_dir
            + "git push -f origin master )"
        )


# fetch gitlab group
# GITLAB_GROUP=$($GITLAB group list | grep -B 1 parlement | head -n 1 | sed 's/.* //')

# echo $GITLAB project create $(head -n 1 .procedure/procedure.csv  | awk -F ';' '{print " --description=\""$2"\" --name='$BILL'"}') --namespace-id=$GITLAB_GROUP --public=true | sh



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys, re, copy

from requests.structures import CaseInsensitiveDict

from tlfp.tools.sort_articles import bister, article_is_lower
from tlfp.tools.common import clean_text_for_diff, compute_similarity, open_json, print_json


def complete(current, previous, step, previous_step_metas, table_concordance=None, anteprevious=None, debug=False):
    current = copy.deepcopy(current)
    previous = copy.deepcopy(previous)
    table_concordance = CaseInsensitiveDict(table_concordance or dict())

    DEBUG = debug or '--debug' in sys.argv
    def log(text):
        if DEBUG:
            print(text, file=sys.stderr)

    def exit():
        raise Exception('[complete_articles] Fatal error')

    find_num = re.compile(r'-?[a-z]*(\d+)(-[ta][\dIV]+|\D)?$')
    oldnum = 0
    oldstep = {}
    oldjson = []
    oldstatus = {}
    oldartids = []
    oldarts = []
    oldsects = []
    olddepot = None
    try:
        for line in previous:
            if line["type"] != "texte":
                oldjson.append(line)
            else:
                oldnum = int(find_num.search(line['id']).group(1))
                olddepot = line['depot']
            if line["type"] == "article":
                keys = list(line['alineas'].keys())
                keys.sort()
                oldstep[line["titre"]] = [line['alineas'][k] for k in keys]
                oldstatus[line["titre"]] = line['statut']
                oldartids.append(line["titre"])
                oldarts.append((line["titre"], line))
            elif line["type"] == "section":
                oldsects.append(line)
    except Exception as e:
        print(type(e), e, file=sys.stderr)
        print("Incorrect previous text: %s" % previous)
        exit()

    if previous_step_metas and not previous_step_metas.get("skip"):
        try:
            assert(previous_step_metas["type"] == "texte")
            oldnum = int(find_num.search(previous_step_metas['id']).group(1))
        except Exception as e:
            print(type(e), e, file=sys.stderr)
            print("Incorrect previous step: %s" % previous_step_metas)
            exit()

    gdoldstep = None
    if anteprevious:
        gdoldstep = {}
        for line in anteprevious:
            if line["type"] == "article":
                keys = list(line['alineas'].keys())
                keys.sort()
                gdoldstep[line["titre"]] = [line['alineas'][k] for k in keys]

    ALL_ARTICLES = []
    def write_json(data):
        nonlocal ALL_ARTICLES
        ALL_ARTICLES.append(data)

    null_reg = re.compile(r'^$')
    re_mat_simple = re.compile(r'[IVXDCLM\d]')
    re_mat_complex = re.compile(r'L[O.\s]*[IVXDCLM\d]')
    re_mat_complex2 = re.compile(r'\d+-\d+')
    re_clean_art = re.compile(r'^"?Art\.?\s*', re.I)
    make_sta_reg = lambda x: re.compile(r'^("?Art[\s\.]*)?%s\s*(([\.°\-]+\s*)+)' % re_clean_art.sub('', x))
    make_end_reg = lambda x, rich: re.compile(r'^%s[IVXDCLM\d\-]+([\-\.\s]+\d*)*((%s|[A-Z])\s*)*(\(|et\s|%s)' % ('("?[LA][LArRtTO\.\s]+)?' if rich else "", bister, x))
    re_sect_chg = re.compile(r'^((chap|t)itre|volume|livre|tome|(sous-)?section)\s+[1-9IVXDC]', re.I)
    def get_mark_from_last(text, s, l="", sep="", enable_copyall=False, copyall=False):
        log("- GET Extract from " + s + " to " + l)
        res = []
        try:
            start = make_sta_reg(s)
        except Exception as e:
            print('ERROR', type(e), e, s, l, file=sys.stderr)
            exit()
        rich = re_mat_complex.match(s) or re_mat_complex2.match(s) or not re_mat_simple.match(s)
        if l:
            last = make_sta_reg(l)
        re_end = None
        record = False
        for n, i in enumerate(text):
            matc = start.match(i)
            # log("    TEST: " + i[:50])
            if re_end and (re_end.match(i) or re_sect_chg.match(i)):
                if l:
                    re_end = make_end_reg(sep, rich)
                    l = ""
                else:
                    log("  --> END FOUND")
                    record = False
                    break
            elif matc:
                sep = matc.group(2).strip()
                log("  --> START FOUND " + sep)
                record = True
                if l:
                    re_end = last
                else:
                    re_end = make_end_reg(sep, rich)
            elif copyall is True:
                record = True
                re_end = null_reg
                if n == 0:
                    i = "%s%s %s" % (s, ". -" if re_mat_simple.match(s) else sep, i)
            if record:
                log("     copy alinea: " + i[:10])
                res.append(i)
        # retry and get everything as I before II added if not found
        if not res:
            if not l and enable_copyall and not copyall:
                log("   nothing found, grabbing all article now...")
                return get_mark_from_last(text, s, l, sep=sep, copyall=True)
            print('ERROR: could not retrieve', s, file=sys.stderr)
            return False
        return res

    get_alineas_text = lambda a: clean_text_for_diff([a[k] for k in sorted(a.keys())])

    re_clean_et = re.compile(r'(\s*[\&,]\s*|\s+et\s+)+', re.I)
    re_suppr = re.compile(r'\W*suppr(ess|im)', re.I)
    re_confo = re.compile(r'\W*(conforme|non[\s\-]*modifi)', re.I)
    re_confo_with_txt = re.compile(r'\s*\(\s*(conforme|non[\s\-]*modifié)\s*\)\s*([\W]*\w+)', re.I)
    re_clean_subsec_space = re.compile(r'^("?[IVX0-9]{1,4}(\s+[a-z]+)?(\s+[A-Z]{1,4})?)\s*([\.°\-]+)\s*([^\s\)])', re.I)
    order = 1
    cursec = {'id': ''}
    done_titre = False
    texte = None
    for line_i, line in enumerate(current):
        if not line or not "type" in line:
            sys.stderr.write("JSON badly formatted, missing field type: %s\n" % line)
            exit()
        if oldnum and 'source_text' in line and oldnum != line['source_text']:
            continue
        if line["type"] == "echec":
            texte["echec"] = True
            texte["expose"] = line["texte"]
            write_json(texte)
            for a in oldjson:
                write_json(a)
            break
        elif line["type"] == "texte":
            texte = dict(line)
            # check number of sections is the same as the final text
            #if texte["definitif"]:
                # assert len(oldsects) == len([x for x in current if x['type'] == 'section'])
        else:
          if not done_titre:
            write_json(texte)
            done_titre = True
          if line["type"] != "article":
            if texte['definitif']:
                try:
                    cursec = oldsects.pop(0)
                    assert(cursec["type_section"] == line["type_section"])
                except:
                    print("ERROR: Problem while renumbering sections: ", line['titre'], " is not ", cursec, '\n', file=sys.stderr)
                    # exit()
                if line["id"] != cursec["id"]:
                    log("DEBUG: Matched section %s (%s) with old section %s (%s)" % (line["id"], line.get('titre'), cursec["id"], cursec.get('titre')))
                    line["newid"] = line["id"]
                    line["id"] = cursec["id"]
            write_json(line)
          else:
            keys = list(line['alineas'].keys())
            keys.sort()
            alineas = [line['alineas'][k] for k in keys]
            mult = line['titre'].split(' à ')
            is_mult = (len(mult) > 1)
            if is_mult:
                st = mult[0].strip()
                ed = mult[1].strip()
                if re_suppr.match(line['statut']) or (len(alineas) == 1 and re_suppr.match(alineas[0])):
                    if (st not in oldartids and ed not in oldartids) or (st in oldstatus and re_suppr.match(oldstatus[st]) and ed in oldstatus and re_suppr.match(oldstatus[ed])):
                        log("DEBUG: SKIP already deleted articles %s to %s" % (st, ed))
                        continue
                    log("DEBUG: Marking as deleted articles %s à %s" % (st, ed))
                    mult_type = "sup"
                elif re_confo.match(line['statut']) or (len(alineas) == 1 and re_confo.match(alineas[0])):
                    log("DEBUG: Recovering art conformes %s à %s" % (st, ed))
                    mult_type = "conf"
                else:
                    print("ERROR: Found multiple article which I don't know what to do with", line['titre'], line, file=sys.stderr)
                    exit()
                line['titre'] = st
            cur = ""
            if texte['definitif']:
                try:
                    # recover the old article and mark those deleted as deleted
                    while True:
                        _, oldart = oldarts[0]

                        # first, mark the old articles as deleted via the concordance table
                        if oldart['titre'].lower() in table_concordance:
                            new_art = table_concordance[oldart['titre']]
                            if 'suppr' in new_art:
                                c, a = oldarts.pop(0)
                                oldartids.remove(c)
                                if olddepot:
                                    log("DEBUG: Marking art %s as supprimé (thanks to concordance table)" % c)
                                    a["order"] = order
                                    order += 1
                                    write_json(a)
                                continue

                        # as a backup, use the detected status to wait for a non-deleted article
                        if re_suppr.match(oldart['statut']):
                            c, a = oldarts.pop(0)
                            oldartids.remove(c)
                            if olddepot:
                                log("DEBUG: Marking art %s as supprimé (recovered)" % c)
                                a["order"] = order
                                order += 1
                                write_json(a)
                        else:
                            break
                except Exception as e:
                    print("ERROR: Problem while renumbering articles", line, "\n", oldart, "\n", type(e), e, file=sys.stderr)
                    exit()

                # detect matching errors
                if oldart['titre'] in table_concordance:
                    new_art = table_concordance[oldart['titre']]
                    if new_art.lower() != line['titre'].lower():
                        print("ERROR: true concordance is different: when parsing article '%s', we matched it with '%s' which should be matched to '%s' (from concordance table) " % (line['titre'] , oldart['titre'], new_art))
                        match = None
                        for oldart_title, newart_title in table_concordance.items():
                            if newart_title.lower() == line['titre'].lower():
                                match = oldart_title
                                print('    -> it should have been matched with article %s' % oldart_title)
                                break
                        else:
                            print('     -> it should have been deleted')

                        # if article not matching but here in the concordance table, introduce it as a new one
                        # since it can correspond to an article deleted by the Senate in Nouvelle lecture
                        # or an article introduced in CMP hémicycle
                        # /!\ this can only happen during a lecture définitive or a CMP hémicycle
                        if (step.get('stage') == 'l. définitive' or (
                                step.get('step') == 'hemicycle' and step.get('stage') == 'CMP')
                            ) and match:
                            log("DEBUG: Marking art %s as nouveau" % line['titre'])
                            if "section" in line and cursec['id'] != line["section"]:
                                line["section"] = cursec["id"]
                            a = line
                            a["order"] = order
                            a["status"] = "nouveau"
                            if a['titre'] != match:
                                a['newtitre'] = a['titre']
                                a['titre'] = match
                            order += 1
                            write_json(a)
                            continue
                        else:
                            exit()

                log("DEBUG: article '%s' matched with old article '%s'" % (line['titre'], oldart['titre']))

                oldtxt = get_alineas_text(oldart["alineas"])
                txt = get_alineas_text(line["alineas"])
                similarity = compute_similarity(oldtxt, txt)
                if similarity < 0.75 and not olddepot and not step.get('stage') == 'constitutionnalité':
                    print("WARNING BIG DIFFERENCE BETWEEN RENUMBERED ARTICLE", oldart["titre"], "<->", line["titre"], len(oldtxt), "Vs", len(txt), "chars, similarity; %.2f" % similarity, file=sys.stderr)

                if line['titre'] != oldart['titre']:
                    line['newtitre'] = line['titre']
                    line['titre'] = oldart['titre']
                if "section" in line and cursec['id'] != line["section"]:
                    line["section"] = cursec["id"]

            if oldarts:
                while oldarts:
                    cur, a = oldarts.pop(0)
                    if line['titre'] in oldartids or article_is_lower(cur, line['titre']):
                        oldartids.remove(cur)
                    else:
                        oldarts.insert(0, (cur, a))
                        break
                    if cur == line['titre']:
                        break

                    if a["statut"].startswith("conforme"):
                        log("DEBUG: Recovering art conforme %s" % cur)
                        a["statut"] = "conforme"
                        a["order"] = order
                        order += 1
                        write_json(a)
                    elif not re_suppr.match(a["statut"]):
                        # if the last line of text was some dots, it means that we should keep
                        # the articles as-is if they are not deleted
                        last_block_was_dots_and_not_an_article = False
                        for block in reversed(current[:line_i]):
                            if block['type'] == 'dots':
                                last_block_was_dots_and_not_an_article = True
                                break
                            if block['type'] == 'article':
                                break
                        if last_block_was_dots_and_not_an_article:
                            # ex: https://www.senat.fr/leg/ppl09-304.html
                            log("DEBUG: Recovering art as non-modifié via dots %s" % cur)
                            a["statut"] = "non modifié"
                            a["order"] = order
                            order += 1
                            write_json(a)
                        else:
                            log("DEBUG: Marking art %s as supprimé because it disappeared" % cur)
                            a["statut"] = "supprimé"
                            a["alineas"] = dict()
                            a["order"] = order
                            order += 1
                            write_json(a)
            if is_mult:
                if ed not in oldartids or cur != line['titre']:
                    if mult_type == "sup":
                        print("WARNING: could not find first or last part of multiple article to be removed:", line['titre'], "to", ed, "(last found:", cur+")", file=sys.stderr)
                        continue
                    print("ERROR: dealing with multiple article", line['titre'], "to", ed, "Could not find first or last part in last step (last found:", cur+")", file=sys.stderr)
                    exit()
                while True:
                    if mult_type == "sup" and not re_suppr.match(a["statut"]):
                        log("DEBUG: Marking art %s as supprimé (mult)" % cur)
                        a["statut"] = "supprimé"
                        a["alineas"] = dict()
                        a["order"] = order
                        order += 1
                        write_json(a)
                    elif mult_type == "conf":
                        log("DEBUG: Recovering art conforme %s (mult)" % cur)
                        a["statut"] = "conforme"
                        a["order"] = order
                        order += 1
                        write_json(a)
                    if cur == ed or not oldarts:
                        break
                    cur, a = oldarts.pop(0)
                continue
            if (re_suppr.match(line["statut"]) or (len(alineas) == 1 and re_suppr.match(alineas[0]))) and (line['titre'] not in oldstatus or re_suppr.match(oldstatus[line['titre']])):
               continue
            # Clean empty articles with only "Non modifié" and include text from previous step
            if alineas and re_confo.match(alineas[0]) and alineas[0].endswith(')'):
                if not line['titre'] in oldstep:
                    sys.stderr.write("WARNING: found repeated article %s missing from previous step: %s (article is ignored)\n" % (line['titre'], line['alineas']))
                    # ignore empty non-modified
                    continue
                else:
                    log("DEBUG: get back Art %s" % line['titre'])
                    alineas = oldstep[line['titre']]
            gd_text = []
            enable_copyall = True
            for j, text in enumerate(alineas):
                if "(Non modifi" in text and not line['titre'] in oldstep:
                    sys.stderr.write("WARNING: found repeated article missing %s from previous step: %s\n" % (line['titre'], text))
                elif re_confo_with_txt.search(text):
                    text = re_confo_with_txt.sub(r' \2', text)
                    text = re_clean_subsec_space.sub(r'\1\4 \5', text)
                    gd_text.append(text)
                elif "(Non modifi" in text:
                    part = re.split("\s*([\.°\-]+\s*)+\s*\(Non", text)
                    if not part:
                        log("ERROR trying to get non-modifiés")
                        exit()
                    pieces = re_clean_et.sub(',', part[0])
                    log("EXTRACT non-modifiés for " + line['titre'] + ": " + pieces)
                    piece = []
                    for todo in pieces.split(','):
                        # Extract series of non-modified subsections of articles from previous version.
                        if " à " in todo:
                            start = re.split(" à ", todo)[0]
                            end = re.split(" à ", todo)[1]
                            mark = get_mark_from_last(oldstep[line['titre']], start, end, sep=part[1:], enable_copyall=enable_copyall)
                            if mark is False and gdoldstep:
                                mark = get_mark_from_last(gdoldstep[line['titre']], start, end, sep=part[1:], enable_copyall=enable_copyall)
                            if mark is False:
                                exit()
                            enable_copyall = False
                            piece.extend(mark)
                        # Extract set of non-modified subsections of articles from previous version.
                        elif todo:
                            mark = get_mark_from_last(oldstep[line['titre']], todo, sep=part[1:], enable_copyall=enable_copyall)
                            if mark is False and gdoldstep:
                                mark = get_mark_from_last(gdoldstep[line['titre']], todo, sep=part[1:], enable_copyall=enable_copyall)
                            if mark is False:
                                exit()
                            enable_copyall = False
                            piece.extend(mark)
                    gd_text.extend(piece)
                else:
                    gd_text.append(text)
            line['alineas'] = dict()
            line['order'] = order
            order += 1
            for i, t in enumerate(gd_text):
                line['alineas']["%03d" % (i+1)] = t
            write_json(line)

    if texte['definitif'] and oldsects and oldarts:
        print("ERROR: %s sections left:\n%s" % (len(oldsects), oldsects), file=sys.stderr)
        #exit()

    while oldarts:
        cur, a = oldarts.pop(0)
        oldartids.remove(cur)
        if texte['definitif'] and not re_suppr.match(a["statut"]):
            print("ERROR: %s articles left:\n%s %s" % (len(oldarts)+1, cur, oldartids), file=sys.stderr)
            exit()

        if not texte.get('echec', '') and a["statut"].startswith("conforme"):
            log("DEBUG: Recovering art conforme %s (leftovers)" % cur)
            a["statut"] = "conforme"
            a["order"] = order
            order += 1
            write_json(a)
        # do not keep already deleted articles but mark as deleted missing ones
        elif not re_suppr.match(a["statut"]) or texte.get('echec', ''):
            # if the last line of text was some dots, it means that we should keep
            # the articles as-is if they are not deleted
            if line['type'] == 'dots':
                # ex: https://www.senat.fr/leg/ppl09-304.html
                log("DEBUG: Recovering art as non-modifié via dots %s (leftovers)" % cur)
                a["statut"] = "non modifié"
                a["order"] = order
                order += 1
                write_json(a)
            else:
                log("DEBUG: Marking art %s as supprimé (leftovers)" % cur)
                a["statut"] = "supprimé"
                a["alineas"] = dict()
                a["order"] = order
                order += 1
                write_json(a)

    return ALL_ARTICLES

if __name__ == '__main__':
    serialized = open_json(sys.argv[1])
    serialized["debug"] = True
    result = complete(**serialized)
    print_json(result)



import re
import sys

from lawfactory_utils.urls import download, enable_requests_cache

from .common import strip_text

# TODO:
# - parse by <p>
# - return ID ECLI
# - return list of signataires saisine
# - return list of signataires décision
# - return list of visa
# - return list of considerants
# - return list of decision articles

re_delibere = re.compile(r"<p>\s*(Jug|Délibér)é par le Conseil constitutionnel .*$", re.S)
clean_delib = lambda x: re_delibere.sub("", x)

def extract_full_decision(url):
    decision_src = download(url).text
    if '<a name=\'visa\' id="visa"></a>' not in decision_src:
        print("ERROR: could not find visa in decision CC", url, file=sys.stderr)
        return None
    decision_txt = decision_src.split('<a name=\'visa\' id="visa"></a>')[1]
    if not re_delibere.search(decision_txt):
        print("ERROR: could not find siège in décision CC", url, file=sys.stderr)
        return None
    decision_txt = clean_delib(decision_txt)
    return strip_text(decision_txt)

def get_decision_length(url):
    decision_txt = extract_full_decision(url)
    if not decision_txt:
        return -1
    return len(decision_txt)

if __name__ == "__main__":
    enable_requests_cache()
    if len(sys.argv) == 2:
        with open(sys.argv[1]) as f:
            for url in f.readlines():
                url = url.strip()
                print(url, ':', get_decision_length(url))
    else:
        print(extract_full_decision(sys.argv[1]))
